"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only


FlowSequencer knows about the DE1 and imports it
DE1 needs to know about the FlowSequencer to be able to set stop-at targets.

Resolve by defining a restricted interface
"""
import enum
import multiprocessing
import time
from typing import Optional

import pyDE1
from pyDE1.de1.c_api import API_MachineStates
from pyDE1.event_manager.payloads import EventPayload
from pyDE1.exceptions import (
    DE1APITypeError, DE1APIValueError, DE1APINotManagedHereException
)


logger = pyDE1.getLogger('FlowSequencer')


class FlowSequencer():
    """
    A singleton base class that helps break circular includes
    Returns the FlowSequencerImpl singleton
    """

    def __new__(cls, *args, **kwds):
        # By importing FlowSequencerImpl at run time,
        # potential import issues should be mitigated.
        import pyDE1.flow_sequencer_impl
        return pyDE1.flow_sequencer_impl.FlowSequencerImpl()

    database_queue: Optional[multiprocessing.Queue] = None
    
    @property
    def stop_at_weight_adjust(self):
        raise NotImplementedError

    @property
    def de1(self):
        raise NotImplementedError

    async def set_up_subscribers(self):
        raise NotImplementedError

    @property
    def scale_processor(self):
        raise NotImplementedError

    @property
    def active_state(self):
        raise NotImplementedError

    def active_control_for_state(self,
                                 state: API_MachineStates) -> 'ByModeControl':
        raise NotImplementedError

    @property
    def active_control(self) -> 'ByModeControl':
        raise NotImplementedError

    @property
    def sequence_start_time(self):
        raise NotImplementedError

    async def stop_at_notify(self, stop_at: 'StopAtType',
                             action: 'StopAtNotificationAction',
                             current: Optional[float] = None):
        raise NotImplementedError

    def stop_at_time_set(self, state: API_MachineStates, duration: float):
        raise NotImplementedError

    def stop_at_volume_set(self, state: API_MachineStates, volume: float):
        raise NotImplementedError

    def stop_at_weight_set(self, state: API_MachineStates, weight: float):
        raise NotImplementedError

    def profile_can_override_stop_limits(self, state: API_MachineStates):
        raise NotImplementedError

    def profile_can_override_tank_temperature(self, state: API_MachineStates):
        raise NotImplementedError

    async def on_de1_nearly_ready(self) -> None:
        raise NotImplementedError


LAST_DROPS_MINIMUM_TIME_DEFAULT = 3.0  # seconds
FIRST_DROPS_THRESHOLD_DEFAULT = 0.0  # bar


class StopAtNotificationAction (enum.Enum):
    ENABLED = 'enabled'
    TRIGGERED = 'triggered'
    DISABLED = 'disabled'
    DE1CONTROLLED = 'de1 controlled'


class StopAtType (enum.Enum):
    TIME = 'time'
    VOLUME = 'volume'
    WEIGHT = 'weight'


class StopAtNotification (EventPayload):
    """
    Enable, disable, trigger of the various stop-at conditions
    current_value is generally only set for StopAtNotificationAction.TRIGGERED

    ENABLED notifications are given even if the target is None as the target
    can be changed during the shot, at least when managed by the FlowSequencer
    (On-the-fly profile changes are not supported at this time.
     On-the-fly changes to steam duration have not been tested at this time.)
    """
    def __init__(self, stop_at: StopAtType,
                 action: StopAtNotificationAction,
                 target_value: Optional[float] = None,
                 current_value: Optional[float] = None,
                 active_state: API_MachineStates = API_MachineStates.NoRequest):
        now = time.time()
        super(StopAtNotification, self).__init__(
            arrival_time=now,
            create_time=now
        )
        self._version = "1.0.0"
        self.stop_at = stop_at
        self.action = action
        self.target_value = target_value
        self.current_value = current_value
        self.active_state = active_state


class AutoTareNotificationAction (enum.Enum):
    ENABLED = 'enabled'
    DISABLED = 'disabled'


class AutoTareNotification (EventPayload):

    def __init__(self, action: AutoTareNotificationAction):
        now = time.time()
        super(AutoTareNotification, self).__init__(
            arrival_time=now,
            create_time=now
        )
        self._version = "1.0.0"
        self.action = action


class ByModeControl:
    """
    Generic holder for stop-at values and other "in-the-moment" parameters
    that are related to flow sequence

    As stop-at-time is handled for steam by the DE1 and that is an async call
    will have to figure out how to manage that at this level. The API already
    directs to the DE1.

    stop_at_time: Steam (DE1), HotWaterRinse, Espresso desirable to add
    stop_at_volume: Espresso, HotWater
    stop_at_weight: Espresso, HotWater
    disable_auto_tare: All

    specials: Espresso only
    """

    def __init__(self, disable_auto_tare: bool = False):
        self._disable_auto_tare = None
        self.disable_auto_tare = disable_auto_tare

    @property
    def disable_auto_tare(self):
        return self._disable_auto_tare

    @disable_auto_tare.setter
    def disable_auto_tare(self, value):
        if not isinstance(value, bool):
            raise DE1APITypeError(
                f"disable_auto_tare must be a bool, not {value}"
            )
        self._disable_auto_tare = value

    @property
    def stop_at_time(self):
        return None

    @property
    def stop_at_weight(self):
        return None

    @property
    def stop_at_volume(self):
        return None

    @property
    def last_drops_minimum_time(self):
        return 0

    @property
    def first_drops_threshold(self):
        return 0

    # Validate these, as they will be coming from the API
    # The API should have already done type validation
    # Though not used by the base class, lowers repetition

    @staticmethod
    def _validate_stop_at(value):
        if value is not None:
            if value == 0:
                value = None
                logger.info(
                    "Deprecated use of 0 for stop-at, replaced by None")
            elif value < 0:
                raise DE1APIValueError(
                    f"Stop-at values need to be non-negative ({value})")
        return value


class StopAtTime (ByModeControl):

    def __init__(self, stop_at_time: Optional[float] = None):
        # Mix-in, call super from concrete instance
        self._stop_at_time = None
        try:
            self.stop_at_time = stop_at_time
        except DE1APINotManagedHereException:
            self._stop_at_time = stop_at_time

    @property
    def stop_at_time(self):
        return self._stop_at_time

    @stop_at_time.setter
    def stop_at_time(self, value):
        self._stop_at_time = self._validate_stop_at(value)


class StopAtVolume (ByModeControl):

    def __init__(self, stop_at_volume: Optional[float] = None):
        # Mix-in, call super from concrete instance
        self._stop_at_volume = None
        self.stop_at_volume = stop_at_volume

    @property
    def stop_at_volume(self):
        return self._stop_at_volume

    @stop_at_volume.setter
    def stop_at_volume(self, value):
        self._stop_at_volume = self._validate_stop_at(value)


class StopAtWeight (ByModeControl):

    def __init__(self, stop_at_weight: Optional[float] = None):
        # Mix-in, call super from concrete instance
        self._stop_at_weight = None
        self.stop_at_weight = stop_at_weight

    @property
    def stop_at_weight(self):
        return self._stop_at_weight

    @stop_at_weight.setter
    def stop_at_weight(self, value):
        self._stop_at_weight = self._validate_stop_at(value)

# Create importable for mode-specific control classes?
# No, just import locally, if needed, from flow_sequencer_impl



"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only

"Null" outbound processor that just keeps the queue clear,
counting the various messages by type,
logging every update_period seconds (and resetting the count)
"""

# Supervise:
#   nothing: loop.add_reader()

# Only import the minimal here, as it potentially ends up in all processes.

import enum
import multiprocessing
import multiprocessing.connection as mpc

import pyDE1.config


class OutboundMode (enum.Enum):
    EventPayload = enum.auto()
    LogRecord = enum.auto()


# See will_topic, below
class MQTTStatusText (enum.Enum):
    on_connection = 'Here'
    on_graceful_disconnect = 'Gone'
    on_will = 'Died'


def run_mqtt_outbound(config: pyDE1.config.Config,
                      log_queue: multiprocessing.Queue,
                      outbound_pipe: mpc.Connection,
                      mode: OutboundMode):

    import asyncio
    import json
    import logging
    import os
    import time

    from collections import Callable
    from socket import gethostname

    import paho.mqtt.client as mqtt
    from paho.mqtt.client import MQTTv5, MQTT_CLEAN_START_FIRST_ONLY

    import pyDE1.pyde1_logging as pyde1_logging
    import pyDE1.shutdown_manager as sm

    from pyDE1.supervise import SupervisedTask

    if mode == OutboundMode.LogRecord:
        logger = pyDE1.getLogger('LogMQTT')
        will_subtopic = 'logging'
    else:
        logger = pyDE1.getLogger('Outbound')
        will_subtopic = 'notification'

    will_topic = "/".join((
        config.mqtt.TOPIC_ROOT, 'status/mqtt', will_subtopic,
    ))

    pyde1_logging.setup_queue_logging(config.logging, log_queue)
    pyde1_logging.config_logger_levels(config.logging)

    client_logger = logger.getChild('MQTTClient')
    client_logger.level = logging.INFO

    # https://github.com/eclipse/paho.mqtt.c/issues/864
    # Add support for Unix-domain sockets #864 (open issue)

    # MQTT_PROTOCOL_VERSION = asyncio_mqtt.client.ProtocolVersion.V5

    loop = asyncio.get_event_loop()
    loop.set_debug(True)

    def on_shutdown_underway_cleanup():
        logger.info("Watching for shutdown event")
        sm.shutdown_underway.wait()
        logger.info("Shutting down MQTT client")
        mqtt_client.loop_stop()
        mqtt_client.disconnect()
        logger.info("Setting cleanup_complete")
        sm.cleanup_complete.set()

    on_shutdown_wait_task = loop.run_in_executor(
        None, on_shutdown_underway_cleanup)

    sm.attach_signal_handler_to_loop(sm.shutdown, loop)

    loop.set_exception_handler(sm.exception_handler)

    def _start_shutdown(sig = None):
        loop.create_task(sm.shutdown(sig, loop))

    async def heartbeat():
        hlog = pyDE1.getLogger(
            f"Heartbeat.{multiprocessing.current_process().name}")
        while not sm.shutdown_underway.is_set():
            await asyncio.sleep(10)
            hlog.debug("===== BEEP =====")

    heartbeat_task = SupervisedTask(heartbeat)

    def on_log_callback(client: mqtt.Client, userdata, level, buf):
        client_logger.info(f"CB: Log: level: {level} '{buf}' ({type(buf)})")

    def on_connect_callback(client, userdata, flags, reasonCode, properties):
        if reasonCode == 0:
            level = logging.INFO
        else:
            level = logging.ERROR
        client_logger.log(level,
            f"CB: Connect: flags: {flags}, reasonCode: {reasonCode}, "
            f"properties {properties}")
        # Split "action" from logging for clarity
        if reasonCode == 0:
            _send_on_connection_status()
        else:
            client_logger.critical(
                f"Connection to MQTT broker failed: {str(reasonCode)}, "
                "initiating process shutdown." )
            loop.call_soon_threadsafe(_start_shutdown)

    def _send_on_connection_status():
        mqtt_client.publish(
            topic=will_topic,
            payload=MQTTStatusText.on_connection.value,
            qos=0,
            retain=True,
        )

    def _send_on_graceful_disconnect():
        logger.info("Publishing graceful disconnect status")
        rc = mqtt_client.publish(
            topic=will_topic,
            payload=MQTTStatusText.on_graceful_disconnect.value,
            qos=0,
            retain=True,
        )
        rc.wait_for_publish()

    class ClientSendsGracefulDisconnect (mqtt.Client):

        def disconnect(self, *args, **kwargs):
            _send_on_graceful_disconnect()
            super(ClientSendsGracefulDisconnect,
                  self).disconnect(*args, **kwargs)

    def on_publish_callback(client, userdata, mid):
        client_logger.debug(f"CB: Published: mid: {mid}")

    # Caught exception in on_disconnect:
    #     on_disconnect_callback() missing 1 required positional argument:
    #         'properties'
    def on_disconnect_callback(client, userdata, reasonCode, properties=None):
        if sm.shutdown_underway.is_set():
            level = logging.INFO
        else:
            level = logging.ERROR
        client_logger.log(level, f"CB: Disconnect: reasonCode: {reasonCode}, "
                                 f"properties {properties}")

    def on_socket_open_callback(client, userdata, socket):
        client_logger.debug(f"CB: Socket open: socket: {socket}")

    def on_socket_close_callback(client, userdata, socket):
        client_logger.debug(f"CB: Socket close: socket: {socket}")

    def on_socket_register_write_callback(client, userdata, socket):
        client_logger.debug(f"CB: Socket register write: socket: {socket}")

    def on_socket_unregister_write_callback(client, userdata, socket):
        client_logger.debug(f"CB: Socket unregister write: socket: {socket}")

    mqtt_client = ClientSendsGracefulDisconnect(
        client_id="{}@{}[{}]".format(
            config.mqtt.CLIENT_ID_PREFIX,
            gethostname(),
            os.getpid(),
        ),
        clean_session=None,  # Required for MQTT5
        userdata=None,
        protocol=MQTTv5,
        transport=config.mqtt.TRANSPORT,
    )

    # mqtt_client.on_log = on_log_callback
    mqtt_client.on_connect = on_connect_callback
    # mqtt_client.on_publish = on_publish_callback
    mqtt_client.on_disconnect = on_disconnect_callback
    mqtt_client.on_socket_open = on_socket_open_callback
    mqtt_client.on_socket_close = on_socket_close_callback
    # mqtt_client.on_socket_register_write = on_socket_register_write_callback
    # mqtt_client.on_socket_unregister_write = on_socket_unregister_write_callback

    mqtt_client.enable_logger(client_logger)

    if config.mqtt.USERNAME is not None:
        logger.info(f"Connecting with username '{config.mqtt.USERNAME}'")
        mqtt_client.username_pw_set(
            username=config.mqtt.USERNAME,
            password=config.mqtt.PASSWORD
        )

    mqtt_client.will_set(topic=will_topic,
                         payload=MQTTStatusText.on_will.value,
                         qos=0,
                         retain=True,
                         properties=None
                         )

    if config.mqtt.TLS:
        mqtt_client.tls_set(ca_certs=config.mqtt.TLS_CA_CERTS,
                            certfile=config.mqtt.TLS_CERTFILE,
                            keyfile=config.mqtt.TLS_KEYFILE,
                            cert_reqs=config.mqtt.TLS_CERT_REQS,
                            tls_version=config.mqtt.TLS_VERSION,
                            ciphers=config.mqtt.TLS_CIPHERS)

    mqtt_client.connect(host=config.mqtt.BROKER_HOSTNAME,
                        port=config.mqtt.BROKER_PORT,
                        keepalive=config.mqtt.KEEPALIVE,
                        bind_address="",
                        bind_port=0,
                        clean_start=MQTT_CLEAN_START_FIRST_ONLY,
                        properties=None)

    mqtt_client.loop_start()

    last_update = time.time()
    update_period = 10  # in seconds
    counts = {}

    def create_pipe_reader_event_payload() -> Callable:

        def outbound_pipe_reader():

            nonlocal last_update, update_period, counts
            nonlocal outbound_pipe, mqtt_client

            item_json = outbound_pipe.recv()
            item_as_dict = json.loads(item_json)
            topic = f"{config.mqtt.TOPIC_ROOT}/{item_as_dict['class']}"
            mqtt_client.publish(
                topic=topic,
                payload=item_json,
                qos=0,
                # retain=True,  # Can cause client to always check if "current"
                retain=False,
                properties=None
            )

            now = time.time()
            try:
                counts[item_as_dict['class']] += 1
            except KeyError:
                counts[item_as_dict['class']] = 1
            if now - last_update > update_period:
                logger.debug(counts)
                counts = {}
                last_update = now

        return outbound_pipe_reader

    def create_pipe_reader_log_record() -> Callable:

        def outbound_pipe_reader():

            nonlocal last_update, update_period, counts
            nonlocal outbound_pipe, mqtt_client

            record: logging.LogRecord = outbound_pipe.recv()

            mqtt_client.publish(
                topic=f"{config.mqtt.TOPIC_ROOT}/log",
                payload=record.getMessage(),
                qos=0,
                retain=False,
                properties=None
            )

            record_as_json = pyde1_logging.log_record_to_json(record)
            mqtt_client.publish(
                topic=f"{config.mqtt.TOPIC_ROOT}/log/record",
                payload=record_as_json,
                qos=0,
                retain=False,
                properties=None
            )

        return outbound_pipe_reader

    if mode == OutboundMode.LogRecord:
        reader = create_pipe_reader_log_record()
    else:
        reader = create_pipe_reader_event_payload()

    loop.add_reader(outbound_pipe.fileno(), reader)

    loop.run_forever()



"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only

Executes in FlowSequencer context, kept in database directory
"""
import json
import time
from typing import Optional

import aiosqlite

from pyDE1.config import config
from pyDE1.de1.c_api import API_MachineStates
from pyDE1.dispatcher.implementation import get_resource_to_dict
from pyDE1.dispatcher.resource import Resource
from pyDE1.event_manager import SequencerGateNotification


async def resource_to_json(resource: Resource):

    resource_dict = await get_resource_to_dict(resource)

    # In HTTP API this gets pretty printed
    # content = json.dumps(content,
    #                      sort_keys=True, indent=4) + "\n"
    # Here we're even farther removed from human eyes
    # and under time pressure

    return json.dumps(resource_dict)


STATE_TO_CONTROL_MAP = {
    API_MachineStates.Espresso: Resource.DE1_CONTROL_ESPRESSO,
    API_MachineStates.Steam: Resource.DE1_CONTROL_STEAM,
    API_MachineStates.HotWater: Resource.DE1_CONTROL_HOT_WATER,
    API_MachineStates.HotWaterRinse: Resource.DE1_CONTROL_HOT_WATER_RINSE
}


async def create_history_record(active_state: API_MachineStates,
                                sequence_start_time: float,
                                profile_id: Optional[str]):
    """
    The main reason to use aiosqlite here is the ability to detect timeout
    In the case of a timeout, the management of flow should continue
    and the recording of the shot should probably stop, both because
    there is no history record, as well as the likelihood of problems
    with writing to the database.

    TODO: Implement that timeout and logic
    """

    t0 = time.time()

    async with aiosqlite.connect(config.database.FILENAME) as db:

        # The database keeps track of the most-recently uploaded profile
        # if there isn't one known by the DE1
        # Happy path: de1 profile id present
        #   Use it
        # Unhappy path: de1 profile null
        #   Option 1 - Query first, include in single commit
        #   Option 2 - INSERT then UPDATE with SELECT and commit

        # https://www.sqlite.org/quirks.html#no_separate_boolean_datatype
        if profile_id is not None:
            profile_assumed = False     # Should match SQLite3 FALSE token
        else:
            profile_assumed = True      # Should match SQLite3 TRUE token
            cur = await db.execute("SELECT value FROM persist_hkv "
                                   "WHERE header == 'last_profile' "
                                   "AND key == 'id'")
            (profile_id,) = await cur.fetchone()

        # Only include fields that have known data

        vals = {
            'id': SequencerGateNotification.sequence_id,
            'active_state': active_state.name,
            'start_sequence': sequence_start_time,
            'profile_id': profile_id,
            'profile_assumed': profile_assumed,
            'resource_version':
                await resource_to_json(Resource.VERSION),
            'resource_de1_id':
                await resource_to_json(Resource.DE1_ID),
            'resource_de1_read_once':
                await resource_to_json(Resource.DE1_READ_ONCE),
            'resource_de1_calibration_flow_multiplier':
                await resource_to_json(
                    Resource.DE1_CALIBRATION_FLOW_MULTIPLIER),
            'resource_de1_control_mode':
                await resource_to_json(STATE_TO_CONTROL_MAP[active_state]),
            'resource_de1_control_tank_water_threshold':
                await resource_to_json(
                    Resource.DE1_CONTROL_TANK_WATER_THRESHOLD),
            'resource_de1_setting_before_flow':
                await resource_to_json(Resource.DE1_SETTING_BEFORE_FLOW),
            'resource_de1_setting_steam':
                await resource_to_json(Resource.DE1_SETTING_STEAM),
            'resource_de1_setting_target_group_temp':
                await resource_to_json(Resource.DE1_SETTING_TARGET_GROUP_TEMP),
            # Confirm that this returns something with no scale
            'resource_scale_id':
                await resource_to_json(Resource.SCALE_ID)
        }

        sql = "INSERT INTO sequence" \
              "(id, active_state, start_sequence, " \
              "profile_id, profile_assumed, " \
              "resource_version, resource_de1_id, resource_de1_read_once, " \
              "resource_de1_calibration_flow_multiplier, " \
              "resource_de1_control_mode, " \
              "resource_de1_control_tank_water_threshold, " \
              "resource_de1_setting_before_flow, " \
              "resource_de1_setting_steam, " \
              "resource_de1_setting_target_group_temp, " \
              "resource_scale_id) " \
              "VALUES " \
              "(:id, :active_state, :start_sequence, " \
              ":profile_id, :profile_assumed, " \
              ":resource_version, :resource_de1_id, " \
              ":resource_de1_read_once, " \
              ":resource_de1_calibration_flow_multiplier, " \
              ":resource_de1_control_mode, " \
              ":resource_de1_control_tank_water_threshold, " \
              ":resource_de1_setting_before_flow, " \
              ":resource_de1_setting_steam, " \
              ":resource_de1_setting_target_group_temp, " \
              ":resource_scale_id)"
        cur = await db.execute(sql, vals)
        await db.commit()


"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only
"""

import asyncio
import json
import multiprocessing
import queue
import threading
import time
from asyncio import Task
from collections import deque
from copy import deepcopy
from typing import Dict, Deque

import aiosqlite

import pyDE1
import pyDE1.database.insert as db_insert
import pyDE1.shutdown_manager as sm
from pyDE1.config import config
from pyDE1.database.recorder_control import RecorderControl

# from pyDE1.dispatcher.dispatcher import QUEUE_TOO_DEEP
QUEUE_TOO_DEEP = 1

from pyDE1.event_manager import SequencerGateName
from pyDE1.event_manager.payloads import EventNotificationAction
from pyDE1.exceptions import DE1TypeError



logger = pyDE1.getLogger('Database.Notifications')


def read_queue_to_queue(queue_to_get: multiprocessing.Queue,
                        queue_to_put: asyncio.Queue,
                        loop: asyncio.AbstractEventLoop,
                        shutdown: threading.Event):
    while not shutdown.is_set():
        try:
            data = queue_to_get.get(timeout=1.0)
            # This needs to be started _after_ the loop is running
            # asyncio.get_running_loop().call_soon_threadsafe(
            loop.call_soon_threadsafe(
                queue_to_put.put_nowait, data)
            if (qd := queue_to_put.qsize()) > QUEUE_TOO_DEEP:
                logger.error(
                    "Notification queue exceeded QUEUE_TOO_DEEP, "
                    f"{qd} > {QUEUE_TOO_DEEP}")
        except queue.Empty:
            pass


async def async_queue_get(from_queue: multiprocessing.Queue):
    loop = asyncio.get_running_loop()
    done = False
    data = None  # For exit on shutdown
    while not done and not sm.shutdown_underway.is_set():
        try:
            # t0 = time.time()
            data = await loop.run_in_executor(
                None,
                from_queue.get, True, 1.0)
                            # blocking, timeout
            # t1 = time.time()
            # logger.info(f"Queue wait time {(t1 - t0)*1000:5.1f} ms")
            done = True
        except queue.Empty:
            pass
    if sm.shutdown_underway.is_set():
        logger.info("Shut down async_queue_get")
    return data


# Target one second of data, or the last-known value
ROLLING_BUFFER_SIZE = {
    'ShotSampleWithVolumesUpdate': 5,   # 4.8 per second on 60 Hz
    'WeightAndFlowUpdate': 10,          # Typically 10 samples per second
    'StateUpdate': 7,                   # Sleep through pour
    'SequencerGateNotification': 16,    # 8 gates to clear and potentially set
    'StopAtNotification': 1,
    'WaterLevelUpdate': 3,              # About 2.5 per second
    'ScaleTareSeen': 3,                 # Limited to 2.5 * period
    'AutoTareNotification': 3,
    'ScaleButtonPress': 3,
    'ConnectivityChange': 8,            # 2 * 4 from disconnected to ready
}
# Total, 59 potential

# If present, only capture (one before and?) all after the sequence begins
ROLLING_BUFFER_TIME_LIMITED = [
    'ShotSampleWithVolumesUpdate',
    'WeightAndFlowUpdate',
    'SequencerGateNotification',
]

# This is probably superfluous, but safer
rolling_buffers_lock = threading.Lock()


# This will likely take a while, run as a task
async def dump_rolling_buffers_to_database(rolling_buffers: Dict[str, Deque],
                                           sequence_id: str,
                                           db: aiosqlite.Connection):

    with rolling_buffers_lock:
        snapshot = deepcopy(rolling_buffers)

    t0 = time.time()
    count = 0
    async with db.cursor() as cur:
        for rb_class, rb in snapshot.items():
            for notification in rb:
                if rb_class == 'SequencerGateNotification' \
                        and notification['sequence_id'] != sequence_id:
                    pass
                else:
                    await db_insert.dict_notification_cursor_only(
                        notification=notification,
                        sequence_id=sequence_id,
                        cur=cur
                    )
                    count += 1
        await db.commit()
    t1 = time.time()
    logger.info(f"Dump of {count} notifications in {(t1-t0)*1000:.3f} ms")


async def record_data(incoming: multiprocessing.Queue):

    # Status:
    #   * Before sequence
    #   * In sequence
    #   * Sequence ended

    WAIT_FOR_SEQUENCE_COMPLETE_TIME = 1.0

    consider_sequence_complete = asyncio.Event()

    # Callback on timeout for waiting for sequence complete packet
    def wait_for_recording_stop_callback(task: Task):
        if isinstance(task.exception(), asyncio.CancelledError):
            logger.warning(
                "Timeout waiting for sequence complete to stop recording")
            consider_sequence_complete.set()
            logger.info(
                "Stopping recording with consider_sequence_complete.set()")
        else:
            pass

    rolling_buffers = {}
    for rb_class, rb in ROLLING_BUFFER_SIZE.items():
        rolling_buffers[rb_class] = deque([], rb)

    async with aiosqlite.connect(config.database.FILENAME) as db:
        try:
            recording = False
            sequence_id = 'dummy'
            waiting_for_id = None
            consider_sequence_complete.set()    # Previous sequence is "done"

            while not sm.shutdown_underway.is_set():
                data = await async_queue_get(incoming)

                if isinstance(data, str):
                    data_dict = json.loads(data)
                    # Always keep the rolling buffers populated
                    # this way there is always pre-history available
                    # and associated with the sequence_id

                    try:
                        with rolling_buffers_lock:
                            rolling_buffers[data_dict['class']].append(
                                data_dict)
                    except KeyError:
                        logger.info("No rolling buffer for "
                                    f"{data_dict['class']}")
                    pass

                    if recording or not consider_sequence_complete.is_set():
                        # The history record has already been created
                        # before the RecorderControl message is sent
                        await db_insert.dict_notification(
                            notification=data_dict,
                            sequence_id=sequence_id,
                            db=db)
                        # Check to see if this is the "matching" sequence complete
                        try:
                            if (not consider_sequence_complete.is_set()
                                    and data_dict['class']
                                    == 'SequencerGateNotification'
                                    and data_dict['name']
                                    == SequencerGateName.GATE_SEQUENCE_COMPLETE.value
                                    and data_dict['action']
                                    == EventNotificationAction.SET.value
                                    and data_dict['sequence_id']
                                    == waiting_for_id):
                                waiting_for_id = None
                                consider_sequence_complete.set()
                                logger.info("The wait is over")
                        except ValueError:
                            pass

                elif isinstance(data, RecorderControl):
                    recording = data.recording
                    sequence_id = data.sequence_id
                    if recording:   # start
                        waiting_for_id = None
                        consider_sequence_complete.clear()
                        logger.info("Starting recorder")
                        # Kick off "dribble" of back data here
                        # Even at 5 ms per INSERT, could be 100 ms or more
                        asyncio.create_task(
                            dump_rolling_buffers_to_database(
                                rolling_buffers=rolling_buffers,
                                sequence_id=sequence_id,
                                db=db,
                            )
                        )
                    else:   # recording stop
                        waiting_for_id = sequence_id
                        # This is raising asyncio.exceptions.CancelledError
                        t_wait = asyncio.create_task(
                            asyncio.wait_for(
                                consider_sequence_complete.wait(),
                                WAIT_FOR_SEQUENCE_COMPLETE_TIME)
                        )
                        t_wait.add_done_callback(
                            wait_for_recording_stop_callback
                        )
                        logger.info("Waiting for sequence_complete packet")

                else:
                    if not sm.shutdown_underway.is_set():
                        raise DE1TypeError(
                            "Unrecognized data type passed for recording:"
                            f"{type(data)}")
                    else:
                        continue

            if sm.shutdown_underway.is_set():
                logger.info("Shut down record_data() loop")

        except asyncio.CancelledError as e:
            logger.info(e)
            await db.close()
            raise





"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only
"""

import asyncio
import time
from typing import Optional, Callable, Coroutine, Union

import aiosqlite
from bleak.backends.device import BLEDevice
from bleak.backends.service import BleakGATTServiceCollection

import pyDE1
import pyDE1.shutdown_manager as sm
from pyDE1.bleak_client_wrapper import BleakClientWrapped
from pyDE1.config import config
from pyDE1.dispatcher.resource import ConnectivityEnum
from pyDE1.event_manager import SubscribedEvent
from pyDE1.event_manager.events import ConnectivityState, ConnectivityChange
from pyDE1.exceptions import DE1APIValueError, DE1NoAddressError
from pyDE1.scale.events import ScaleWeightUpdate, ScaleTareSeen
from pyDE1.scanner import _registered_ble_prefixes

logger = pyDE1.getLogger('Scale')

# Used for factory and for BLE detection and filtering
_prefix_to_constructor = dict()
_recognized_scale_prefixes = set()


def recognized_scale_prefixes():
    return _recognized_scale_prefixes.copy()

# TODO: Experimentaly confirm that weight and mass-flow estimates
#       are reasonably time aligned - NB: DE1.fall_time

class Scale:

    def __init__(self):
        self._address_or_bledevice: Optional[Union[str, BLEDevice]] = None
        self._name: Optional[str] = None

        # These are often model-specific, override in subclass init
        self._nominal_period = 0.1  # seconds per sample
        self._minimum_tare_request_interval = 2.5 * self._nominal_period
        self._sensor_lag = 0.38  # seconds, including all delays to arrival
        # From https://www.youtube.com/watch?v=SIzFhnZ32Y0
        # (James Hoffmann) at 4:51
        #   Hiroia    0.20
        #   Skale     0.33
        #   Felicita  0.45
        #   Acaia     0.64
        self._tare_timeout = 1.0  # seconds until considered coincidence
        self._tare_threshold = 0.05  # grams, within this, considered "at zero"
        self.hold_at_tare = False

        self._bleak_client: Optional[BleakClientWrapped] = None

        self._event_connectivity: SubscribedEvent = SubscribedEvent(self)
        self._event_weight_update: SubscribedEvent = SubscribedEvent(self)
        self._event_button_press: SubscribedEvent = SubscribedEvent(self)
        self._event_tare_seen: SubscribedEvent = SubscribedEvent(self)

        self._ready = asyncio.Event()

        self._reconnect_count = 0
        self._log_reconnect_attempts = True

        asyncio.get_event_loop().create_task(
            self._event_connectivity.publish(
                self._connectivity_change(
                    arrival_time=time.time(),
                    state=ConnectivityState.NOT_READY)))

        self._estimated_period = self._nominal_period
        self._last_weight_update_received = 0
        self._last_tare_request_sent = 0

        # See Scale.decommission()
        self._to_decommission = (
            '_event_connectivity',
            '_event_weight_update',
            '_event_button_press',
            '_event_tare_seen',
        )

        # TODO: Think about how to manage a "tare seen" event
        #       and what the use cases for it would be.
        #       Could use asyncio.Event(), but what are the states
        #       and how do you "release" if it never arrives?

        self._tare_requested = False
        self._period_estimator = PeriodEstimator(self)

        # Don't need to await this on instantiation
        asyncio.get_event_loop().create_task(
            self._event_weight_update.subscribe(self._create_self_callback()))


    @property
    def type(self):
        return self.__class__.__name__

    @property
    def address(self):
        addr = self._address_or_bledevice
        if isinstance(addr, BLEDevice):
            addr = addr.address
        return addr

    @address.setter
    def address(self, address: Union[BLEDevice, str]):
        if self.address is not None:
            raise DE1APIValueError(
                "Changing the Scale address is not yet supported")
        self._address_or_bledevice = address
        if isinstance(address, BLEDevice):
            self._name = address.name
        self._bleak_client = BleakClientWrapped(self._address_or_bledevice)

    @property
    def name(self):
        return self._name

    @property
    def sensor_lag(self):
        return self._sensor_lag

    async def connect(self, timeout: Optional[float] = None):

        if timeout is None:
            timeout = config.bluetooth.CONNECT_TIMEOUT

        class_name = type(self).__name__
        logger.info(f"Connecting to {class_name} at {self.address}")

        assert self._bleak_client is not None

        if not self.is_connected:

            self._bleak_client.set_disconnected_callback(
                self._create_disconnect_callback()
            )

            await asyncio.gather(self._event_connectivity.publish(
                self._connectivity_change(
                    arrival_time=time.time(),
                    state=ConnectivityState.CONNECTING)),
                self._bleak_client.connect(timeout=timeout),
                return_exceptions=True
            )

            if self.is_connected:
                self._address_or_bledevice = self._bleak_client.address
                if self.name is None:
                    self._name = self._bleak_client.name
                logger.info(f"Connected to {class_name} at {self.address}")
                await self._event_connectivity.publish(
                    self._connectivity_change(
                        arrival_time=time.time(),
                        state=ConnectivityState.CONNECTED))
                # This can take some time, potentially delaying DE1 connection
                # At least BlueZ doesn't like concurrent connection requests
                asyncio.create_task(self.standard_initialization())

            else:
                logger.error(
                    f"Connection failed to {class_name} at {self.address}")
                await self._notify_not_ready()
                await self._event_connectivity.publish(
                    self._connectivity_change(
                        arrival_time=time.time(),
                        state=ConnectivityState.DISCONNECTED))

    async def standard_initialization(self, hold_notification=False):
        """
        :param hold_notification: Since subclass may need to do more
        """
        logger.info("Scale.standard_initialization()")
        await self.display_on()
        await self.start_sending_weight_updates()
        if self.supports_button_press:
            await self.start_sending_button_updates()
        await self._restore_period_from_db()
        if not hold_notification:
            await self._notify_ready()

    async def _notify_ready(self):
        self._ready.set()
        await self._event_connectivity.publish(
            self._connectivity_change(
                arrival_time=time.time(),
                state=ConnectivityState.READY))
        logger.info("Ready")

    async def _notify_not_ready(self):
        self._ready.clear()
        await self._event_connectivity.publish(
            self._connectivity_change(
                arrival_time=time.time(),
                state=ConnectivityState.NOT_READY))

    @property
    def is_ready(self):
        return self._ready.is_set()

    # Helper method to populate a ConnectivityChange

    def _connectivity_change(self, arrival_time: float,
                             state: ConnectivityState):
        return ConnectivityChange(arrival_time=arrival_time,
                                  state=state,
                                  id=self.address,
                                  name=self.name)

    async def disconnect(self):
        class_name = type(self).__name__
        logger.info(f"Disconnecting from {class_name}")
        if self._bleak_client is None:
            logger.info(f"Disconnecting from {class_name}, no client")
            return

        if self.is_connected:
            await asyncio.gather(
                self._bleak_client.disconnect(),
                self._notify_not_ready(),
                self._event_connectivity.publish(
                    self._connectivity_change(
                        arrival_time=time.time(),
                        state=ConnectivityState.DISCONNECTING)),
                return_exceptions=True
            )

        if self.is_connected:
            logger.error(
                f"Disconnect failed from {class_name} at {self.address}")
            await self._event_connectivity.publish(
                self._connectivity_change(
                    arrival_time=time.time(),
                    state=ConnectivityState.CONNECTED))
        else:
            logger.info(
                f"Scale.disconnect(): Disconnected from {class_name} "
                f"at {self.address}")
            await self._event_connectivity.publish(
                self._connectivity_change(
                    arrival_time=time.time(),
                    state=ConnectivityState.DISCONNECTED))

    @property
    def is_connected(self):
        if self._bleak_client is None:
            return False
        else:
            return self._bleak_client.is_connected

    async def start_sending_weight_updates(self):
        raise NotImplementedError

    async def stop_sending_weight_updates(self):
        raise NotImplementedError

    @property
    def is_sending_weight_updates(self):
        raise NotImplementedError

    @property
    def supports_button_press(self):
        return False

    async def start_sending_button_updates(self):
        raise NotImplementedError

    async def stop_sending_button_updates(self):
        raise NotImplementedError

    async def tare(self):
        """
        A tare request can only be made every
        self._minimum_tare_request_interval seconds

        It doesn't make sense to hammer it as it will take
        at least one reporting period to "see" the tare
        """
        dt = time.time() - self._last_tare_request_sent
        if dt > self._minimum_tare_request_interval:
            await self._tare_internal()
            self._last_tare_request_sent = time.time()
            self._tare_requested = True
            logger.info(f"Tare request sent")
        else:
            logger.info(
                f"Tare request skipped, too soon, {dt:0.3f} seconds")
        return self._last_tare_request_sent


    async def _tare_internal(self):
        raise NotImplementedError

    async def current_weight(self):
        raise NotImplementedError

    async def display_on(self):
        raise NotImplementedError

    async def display_off(self):
        raise NotImplementedError

    # The two *_bool for API

    async def tare_with_bool(self, do_it=True):
        if do_it:
            await self.tare()

    async def display_bool(self, on: bool):
        if on:
            await self.display_on()
        else:
            await self.display_off()

    @property
    def estimated_period(self):
        return self._estimated_period

    @property
    def event_weight_update(self):
        return self._event_weight_update

    @property
    def event_button_press(self):
        return self._event_button_press

    @property
    def event_tare_seen(self):
        return self._event_tare_seen

    def _scale_time_from_latest_arrival(self,
                                        latest_arrival: float):
        """
        Given the latest arrival, provide "best" estimate
        of when that weight was on the scale

        At present, just compensates for scale._scale_delay
        which should include transit delays and the like
        """
        return latest_arrival - self._sensor_lag

    def _update_scale_time_estimator(self,
                                     latest_arrival:float):
        """
        Call once per arrival to update any "fancy" algorithms such as PLL
        """
        pass

    def _create_self_callback(self) -> Coroutine:
        scale = self

        async def self_callback(swu: ScaleWeightUpdate) -> None:
            nonlocal scale
            dt = swu.arrival_time - scale._last_weight_update_received
            scale._last_weight_update_received = swu.arrival_time

            # TODO: Run profiler and evaluate if creating a task
            #       is consuming too much time

            asyncio.create_task(
                scale._period_estimator.process_arrival(dt))

            if scale._tare_requested:
                dt = swu.arrival_time - scale._last_tare_request_sent
                if dt > scale._tare_timeout:
                    scale._tare_requested = False
                    logger.error(f"No tare seen after {dt:0.03f} seconds")
                elif abs(swu.weight) < scale._tare_threshold:
                    scale._tare_requested = False
                    await scale.event_tare_seen.publish(
                        ScaleTareSeen(swu.arrival_time)
                    )
                    logger.info(f"Tare seen after {dt:0.03f} seconds")

            if scale.hold_at_tare:
                if abs(swu.weight) > scale._tare_threshold:
                    # Timing will be checked in scale.tare()
                    await scale.tare()

        return self_callback

    def decommission(self):
        # A Scale has several self-references that may prevent GC
        # Rather than try to deal with weakref, just remove the callbacks
        # Oh, and the list of those collbacks!
        logger.info(f'Decommissioning {self} at {self.address}')

        self._bleak_client.set_disconnected_callback(None)
        for break_ref in self._to_decommission:
            setattr(self, break_ref, None)
        self._to_decommission = None

        # logger.debug("Before GC")
        # for ref in gc.get_referrers(self):
        #     logger.info(f"0x{id(self):x} <== {ref}")
        #
        # logger.info(f"0x{id(self):x} is_finalized: {gc.is_finalized(self)}")

    #  disconnected_callback (callable): Callback that will be scheduled in the
    #  event loop when the client is disconnected. The callable must take one
    #  argument, which will be this client object.

    # The callback seems to be expected to be a "plain" function

    def _create_disconnect_callback(self) -> Callable:
        scale = self

        def disconnect_callback(client: BleakClientWrapped):
            nonlocal scale
            class_name = type(self).__name__
            logger.info(
                "disconnect_callback: "
                f"Disconnected from {class_name} at {client.address}, "
                "willful_disconnect: "
                f"{client.willful_disconnect}")
            asyncio.ensure_future(asyncio.gather(
                self._notify_not_ready(),
                scale._event_connectivity.publish(
                    self._connectivity_change(
                        arrival_time=time.time(),
                        state=ConnectivityState.DISCONNECTED)),
                return_exceptions=True)
            )

            if not client.willful_disconnect:
                asyncio.get_event_loop().create_task(self._reconnect())

        return disconnect_callback

    def _reset_reconnect(self):
        self._reconnect_count = 0
        self._log_reconnect_attempts = True

    async def _reconnect(self):
        """
        Will try immediately config.bluetooth.RECONNECT_RETRY_COUNT times
            of CONNECT_TIMEOUT each. It will then stop logging and retry
            with RECONNECT_GAP seconds in between attempts
        """

        if sm.shutdown_underway.is_set():
            return

        if self._reconnect_count >= config.bluetooth.RECONNECT_RETRY_COUNT:
            await asyncio.sleep(config.bluetooth.RECONNECT_GAP)

        if self._log_reconnect_attempts:
            class_name = type(self).__name__
            logger.info(
                f"Will try reconnecting to {class_name} at {self.address}")

        # Workaround for https://github.com/hbldh/bleak/issues/376
        self._bleak_client.services = BleakGATTServiceCollection()

        await self.connect()
        if self.is_connected:
            self._reset_reconnect()

        else:
            if self._reconnect_count <= config.bluetooth.RECONNECT_RETRY_COUNT:
                self._reconnect_count = self._reconnect_count + 1
            if self._reconnect_count == config.bluetooth.RECONNECT_RETRY_COUNT:
                logger.info("Suppressing further reconnect messages. "
                            "Will keep trying at {}-second intervals.".format(
                    config.bluetooth.RECONNECT_GAP)
                )
                self._log_reconnect_attempts = False
            asyncio.get_event_loop().create_task(
                self._reconnect(), name='ReconnectScale')

    @property
    def nominal_period(self):
        return self._nominal_period

    @nominal_period.setter
    def nominal_period(self, value):
        self._nominal_period = value
        self._period_estimator.reset(self._nominal_period)

    async def _persist_period_to_db(self):
        if not self.address:
            raise DE1NoAddressError(
                "Can't persist scale period without a scale address")
        async with aiosqlite.connect(config.database.FILENAME) as db:
            sql = "INSERT OR REPLACE INTO persist_hkv " \
                  "(header, key, value) " \
                  "VALUES " \
                  "(:header, :key, :value) "
            await db.execute(sql, {
                'header': 'scale.period',
                'key': self.address,
                'value': self.nominal_period,
            })
            await db.commit()

    async def _restore_period_from_db(self):
        if not self.address:
            raise DE1NoAddressError(
                "Can't restore scale period without a scale address")
        async with aiosqlite.connect(config.database.FILENAME) as db:
            sql = "SELECT value FROM persist_hkv " \
                  "WHERE header = :header AND key = :key"
            cur = await db.execute(sql, {
                'header': 'scale.period',
                'key': self.address,
            })
            row = await cur.fetchone()
            if row and row[0]:
                val = float(row[0])
                logger.info(
                    "Loading scale-period estimate of "
                    f"{val:.5f} from database")
                self._estimated_period = val
                self._period_estimator.reset(val)
            else:
                logger.info(
                    "No previous scale-period estimate for "
                    f"{self.address} found")

    # For API
    @property
    def connectivity(self):
        retval = ConnectivityEnum.NOT_CONNECTED
        if self.is_connected:
            if self._ready.is_set():
                retval = ConnectivityEnum.READY
            else:
                retval = ConnectivityEnum.CONNECTED
        return retval

    @staticmethod
    def register_constructor(constructor: Callable, prefix: str):
        _prefix_to_constructor[prefix] = constructor
        _recognized_scale_prefixes.add(prefix)
        _registered_ble_prefixes.add(prefix)


# Used by Scale instances
class PeriodEstimator:
    """
    Estimate inter-arrival period from stream of arrivals

    Presently just an exponential moving average

    Skale II usually "bulks up" two or more reports on a 150-ms clock
    300 ms burbles aren't uncommon. A "normal" 50-ms stretch before its
    other half arrives would generate a (50/100) * k change.
    The other half then would generate (-100/100) * k change
    So k on the order of 1/1000 should be reasonable (10 sec, ~1 min settle)
    k of 1/10000 would be even better (100 sec, 10 min settle)
    Another way to look at this is 600 ms error / 600 s measurement ~ 0.1%

    Hand-in-hand with this is how long to consider a gap vs. a burble
    Nearly 5% of 150-ms windows from a SkaleII had 3 reports.
    Up to 6 in a window were observed. It dropped to 0.1% at 4 reports
    per window. Ignoring too many of these can lead to the estimate being off.
    Based on this, 300 ms (two periods) seems too short.
    300 + 150/2 = 375 ms is probably OK.
    450 + 150/2 = 525 ms is probablu conservative
    Try 500 ms to be reasonable.
    """

    def __init__(self, my_scale):

        # TODO: How to update this PeriodEstimator for subclass changes?

        self._scale = my_scale

        self._k = 1/10000   # tau ~ 17 min at 10 samples/sec
        self._ma = self._scale.nominal_period
        self._too_long = 0.5  # seconds before considered a gap

        self._persist_every_n = 1000  # about 100 seconds
        self._n_counter = 0

    def reset(self, nominal_period: float):
        self._ma = nominal_period

    async def process_arrival(self, delta_arrival_time: float):

        if delta_arrival_time < self._too_long:
            self._ma = ((1 - self._k) * self._ma) \
                       + (self._k * delta_arrival_time)
            self._scale.nominal_period = self._ma
            self._n_counter += 1
            if self._n_counter >= self._persist_every_n:
                self._n_counter = 0
                logger.debug(f"Scale period: {self._ma}")
                await self._scale._persist_period_to_db()



def scale_factory(ble_device: BLEDevice)-> Scale:
    constructor = None
    try:
        constructor = _prefix_to_constructor[ble_device.name]
    except KeyError:
        for prefix in _prefix_to_constructor.values():
            if ble_device.name.startswith(prefix):
                constructor = _prefix_to_constructor[prefix]
    if constructor is None:
        raise DE1APIValueError(
            f"No recognized scale registered for {ble_device.name}"
        )
    logger.debug(f"Creating a new instance of {constructor} "
                 f"from {ble_device}")
    scale: Scale = constructor()
    scale.address = ble_device
    return scale





"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only
"""
import asyncio
import functools
import json
import logging
import logging.handlers
import os
import os.path
import queue

from pprint import pformat
from socket import gethostname
from typing import Optional, NamedTuple, Callable

import aiosqlite
import paho.mqtt.client as mqtt
from paho.mqtt.client import MQTTMessage, MQTTv5, MQTT_CLEAN_START_FIRST_ONLY

import requests
from requests.auth import HTTPBasicAuth, HTTPDigestAuth

from pyDE1.config_load import ConfigYAML, ConfigLoadable

import pyDE1.shutdown_manager as sm
from pyDE1.de1.c_api import API_MachineStates, API_Substates
from pyDE1.event_manager import SequencerGateName
from pyDE1.event_manager.payloads import EventNotificationAction
from pyDE1.exceptions import DE1IncompleteSequenceRecordError
from pyDE1.shot_file.legacy import legacy_shot_file
# from pyDE1.database.write_notifications import async_queue_get

# The default config file can be missing without error
# If specified on the command line and missing is fatal

import pyDE1
import pyDE1.pyde1_logging as pyde1_logging
from pyDE1.pyde1_logging import ConfigLogging


class Config (ConfigYAML):

    DEFAULT_CONFIG_FILE = '/usr/local/etc/pyde1/pyde1-visualizer.conf'

    def __init__(self):
        super(Config, self).__init__()
        self.database = _Database()
        self.logging = _Logging()
        self.mqtt = _MQTT()
        self.visualizer = _Visualizer()


    # This craziness is so pyCharm autocompletes
    # Otherwise typing.SimpleNamespace() would be sufficient

class _MQTT (ConfigLoadable):
    def __init__(self):
        self.TOPIC_ROOT = 'pyDE1'
        self.CLIENT_ID_PREFIX = 'pyde1-visualizer'
        self.BROKER_HOSTNAME = '::1'
        self.BROKER_PORT = 1883
        self.TRANSPORT = 'tcp'
        self.KEEPALIVE = 60
        self.USERNAME = None
        self.PASSWORD = None
        self.DEBUG = False
        self.TLS = False    # Set True, or rest of TLS is ignored
        # See paho Client.tls_set() for details
        self.TLS_CA_CERTS = None
        self.TLS_CERTFILE = None
        self.TLS_KEYFILE = None
        self.TLS_CERT_REQS = None
        self.TLS_VERSION = None
        self.TLS_CIPHERS = None


class _Visualizer (ConfigLoadable):
    def __init__(self):
        self.USERNAME = 'you@example.com'
        self.PASSWORD = 'your password or upload token here'
        self.MIN_FLOW_TIME = 10  # seconds, or not uploaded


class _Logging (ConfigLogging):
    def __init__(self):
        super(_Logging, self).__init__()
        # NB: The log file name is matched against [a-zA-Z0-9._-]
        self.LOG_FILENAME = 'visualizer.log'
        self.LOGGERS = {
            'MQTTClient':       'INFO',
            'root.aiosqlite':   'INFO',
            'root.asyncio':     'INFO',
        }
        self.formatters.STYLE = '%'
        self.formatters.LOGFILE = \
            '%(asctime)s %(levelname)s %(name)s: %(message)s'
        self.formatters.STDERR = \
                        '%(levelname)s %(name)s: %(message)s'

# Accept defaults for other logging


class _Database (ConfigLoadable):
    def __init__(self):
        self.FILENAME = '/var/lib/pyde1/pyde1.sqlite3'


config = Config()


#
# "Temporarily" repeated and modified here
#
async def async_queue_get(from_queue: queue.Queue):
    loop = asyncio.get_running_loop()
    done = False
    data = None  # For exit on shutdown
    while not done and not sm.shutdown_underway.is_set():
        try:
            data = await loop.run_in_executor(
                None,
                from_queue.get, True, 1.0)
                            # blocking, timeout
            done = True
        except queue.Empty:
            pass
    if sm.shutdown_underway.is_set():
        logger.info("Shut down async_queue_get")
    return data


shot_complete_queue = queue.Queue()


class ShotCompleteItem (NamedTuple):
    sequence_id:    str
    flow_start:     float
    pour_start:     float
    flow_end:       float


def report_upload(sci: ShotCompleteItem,
                  url: Optional[str],
                  success: bool,
                  client: mqtt.Client):

    topic_upload = f"{config.mqtt.TOPIC_ROOT}/VisualizerUpload"

    payload = sci._asdict()
    payload['url'] = url
    payload['success'] = success

    client.publish(
        topic=topic_upload,
        payload=json.dumps(payload),
        retain=True
    )


def configure_mqtt() -> mqtt.Client:
    """
    Configure the MQTT client, start it, and return it
    so that it can be used to publish on upload result
    (returns with MQTT client loop running in executor)
    """

    userdata_dict = {
        "sequence_id": None,
        "in_sequence": False,
        "flow_start": 0,
        "pour_start": 0,
        "flow_end": 0,
    }

    # Set up the MQTT client

    topic_states = f"{config.mqtt.TOPIC_ROOT}/StateUpdate"
    topic_gates = f"{config.mqtt.TOPIC_ROOT}/SequencerGateNotification"

    def on_connect_callback(client, userdata, flags, reasonCode, properties):
        logger_mqtt.debug(f"MQTT Connect: flags: {flags}, "
                          f"reasonCode: {reasonCode}, properties {properties}")
        client.subscribe(topic_states)
        client.subscribe(topic_gates)

    def on_message_callback(client: mqtt.Client, userdata,
                            message: MQTTMessage):

        payload_dict = json.loads(message.payload)

        logger_mqtt.debug(f"Received: {message.topic} {message.payload}")
        if message.topic == topic_gates:
            sequence_id = payload_dict['sequence_id']
            name = payload_dict['name']
            action = payload_dict['action']
            active_state = payload_dict['active_state']
            event_time = payload_dict['event_time']

            if action == EventNotificationAction.CLEAR.value:
                pass

            # Note: Gate notifications slightly lag state-change times
            #       as they are the time of Event().set(),
            #       and don't take the time from another notification

            elif not userdata['in_sequence']:
                if name == SequencerGateName.GATE_SEQUENCE_START.value \
                        and active_state == API_MachineStates.Espresso.name:
                    userdata['in_sequence'] = True
                    userdata['sequence_id'] = sequence_id
                    userdata['flow_start'] = 0
                    userdata['pour_start'] = 0
                    userdata['flow_end'] = 0
                    logger.info(f"Sequence started {sequence_id}")
                else:
                    pass    # Assumes gates arrive in order

            else:   # in a sequence

                if sequence_id != userdata['sequence_id']:
                    logger.error("Unexpected sequence id: "
                              f"got {id}, aborting sequence")
                    userdata['in_sequence'] = False

                if name == SequencerGateName.GATE_FLOW_BEGIN.value:
                    userdata['flow_start'] = event_time
                    logger.info('Flow started')

                elif name == SequencerGateName.GATE_FLOW_END.value:
                    userdata['flow_end'] = event_time
                    logger.info('Flow ended')

                elif name == SequencerGateName.GATE_SEQUENCE_COMPLETE.value:
                    userdata['in_sequence'] = False
                    logger_upload.info('Upload triggered')
                    shot_complete_queue.put_nowait(
                        ShotCompleteItem(
                            sequence_id=userdata['sequence_id'],
                            flow_start=userdata['flow_start'],
                            pour_start=userdata['pour_start'],
                            flow_end=userdata['flow_end'],
                        )
                    )
                    t_pi = userdata['pour_start'] - userdata['flow_start']
                    t_pour = userdata['flow_end'] - userdata['pour_start']
                    t_total = userdata['flow_end'] - userdata['flow_start']
                    logger.debug(
                        f"Timing: {t_pi:.0f} + {t_pour:.0f} for "
                        f"{t_total:.0f} seconds"
                    )

                else:
                    pass

        elif message.topic == topic_states:

            # Don't really need this, but will for GUI to split
            # preinfuse and flow time for display

            if (userdata['in_sequence']
                    and payload_dict['previous_substate']
                        != API_Substates.Pour.name
                    and payload_dict['substate']
                        == API_Substates.Pour.name):
                userdata['pour_start'] = payload_dict['event_time']
                logger.info('Pour start')

        else:
            pass

    mqtt_client = mqtt.Client(
        client_id="{}@{}[{}]".format(
            config.mqtt.CLIENT_ID_PREFIX,
            gethostname(),
            os.getpid(),
        ),
        clean_session=None,  # Required for MQTT5
        userdata=userdata_dict,
        protocol=MQTTv5,
        transport=config.mqtt.TRANSPORT,
    )

    if config.mqtt.TLS:
        mqtt_client.tls_set(ca_certs=config.mqtt.TLS_CA_CERTS,
                            certfile=config.mqtt.TLS_CERTFILE,
                            keyfile=config.mqtt.TLS_KEYFILE,
                            cert_reqs=config.mqtt.TLS_CERT_REQS,
                            tls_version=config.mqtt.TLS_VERSION,
                            ciphers=config.mqtt.TLS_CIPHERS)

    if config.mqtt.USERNAME is not None:
        logger.info(f"Connecting with username '{config.mqtt.USERNAME}'")
        mqtt_client.username_pw_set(
            username=config.mqtt.USERNAME,
            password=config.mqtt.PASSWORD
        )

    if config.mqtt.DEBUG:
        paho_logger = pyDE1.getLogger('paho')
        paho_logger.setLevel(logging.DEBUG)
        mqtt_client.enable_logger(paho_logger)

    mqtt_client.on_connect = on_connect_callback
    mqtt_client.on_message = on_message_callback

    return mqtt_client


def run_mqtt_client_sync(mqtt_client: mqtt.Client):

    mqtt_client.connect(host=config.mqtt.BROKER_HOSTNAME,
                   port=config.mqtt.BROKER_PORT,
                   keepalive=config.mqtt.KEEPALIVE,
                   bind_address="",
                   bind_port=0,
                   clean_start=MQTT_CLEAN_START_FIRST_ONLY,
                   properties=None)

    # "start" returns immediately
    # mqtt_client.loop_start()
    # "forever" returns on exception
    # No other reasonable way to catch, see
    # https://stackoverflow.com/questions/2829329/
    #   catch-a-threads-exception-in-the-caller-thread
    mqtt_client.loop_forever()


async def wait_then_cleanup(client: mqtt.Client):
    logger.debug("Waiting for shutdown_underway to shutdown MQTT")
    await loop.run_in_executor(None, sm.shutdown_underway.wait)
    logger.info("shutdown_underway seen as set")
    client.disconnect()
    logger.info("mqtt.Client.disconnect returned")
    # As of 2021-08, loop_stop() will block and force= is ignored
    client.loop_stop()
    logger.info("mqtt.Client.loop_stop returned, setting cleanup_complete")
    sm.cleanup_complete.set()


async def loop_on_queue(client: mqtt.Client):

    async with aiosqlite.connect(config.database.FILENAME) as db:
        while not sm.shutdown_underway.is_set():
            logger.info("Ready and waiting")
            got: ShotCompleteItem = await async_queue_get(
                from_queue=shot_complete_queue)
            # None gets returned on termination of async_queue_get()
            if got is None:
                if not sm.shutdown_underway.is_set():
                    raise RuntimeError(
                        "async_queue_get() unexpectedly returned None")
                else:
                    break

            if config.visualizer.MIN_FLOW_TIME \
                    and (dt := got.flow_end - got.flow_start) \
                            < config.visualizer.MIN_FLOW_TIME:
                logger.info(f"Not uploading, too short {dt:.1f} < "
                            f"{config.visualizer.MIN_FLOW_TIME} seconds")
                continue

            sid = got.sequence_id

            # As the database is processing these packets real-time as well
            # it is possible that the sequence row hasn't been updated yet
            # This will result in DE1DBIncompleteSequenceRecord being raised

            contents = None
            for cnt in range(0,10):
                try:
                    await asyncio.sleep(0.100)  # 100 ms
                    contents = await legacy_shot_file(sequence_id=sid, db=db)
                    break
                except DE1IncompleteSequenceRecordError as e:
                    logger.debug(e)
            if contents is None:
                logger.error("Did not recover, aborting upload: {e}")
                report_upload(sci=got, url=None, success=False,
                              client=client)
                continue

            fd = {
                'file': contents,
                'filename': f"{sid}.shot"
            }

            r = await asyncio.get_running_loop().run_in_executor(
                None,
                functools.partial(
                    requests.post,
                    url='https://visualizer.coffee/api/shots/upload',
                    files=fd,
                    auth=HTTPBasicAuth(username=config.visualizer.USERNAME,
                                       password=config.visualizer.PASSWORD)
                ))
            r: requests.Response
            if r.ok:
                d = json.loads(r.text)
                url = f"https:///visualizer.coffee/shots/{d['id']}"
                logger_upload.info(url)
                report_upload(sci=got, url=url, success=True,
                              client=client)
            else:
                logger_upload.error(
                    f"Upload failed: {r.status_code} {r.reason} {r.text}")
                logger_upload.debug(pformat(r))
                report_upload(sci=got, url=None, success=False,
                              client=client)


async def setup_and_run():
    global client   # TODO: This is sort of hack-ish
    client = configure_mqtt()
    loop.create_task(wait_then_cleanup(client))
    mqtt_task = loop.run_in_executor(None, run_mqtt_client_sync, client)
    mqtt_task.add_done_callback(sm.shutdown_if_exception)
    logger.info(f"MQTT run_in_executor started")
    # await asyncio.gather(
    #     mqtt_task,
    await loop_on_queue(client=client)
    #     return_exceptions=True
    # )


if __name__ == '__main__':

    import argparse
    ap = argparse.ArgumentParser(
        description="""Service to upload pyDE1 "shots" to visualizer.coffee
        
        Listens to MQTT announcements of flow and state from pyDE1. 
        When complete, accesses the (local) database for the "shot file" 
        and uploads to visualizer.coffee as well as notifying with URL on 
        {config.mqtt.TOPIC_ROOT}/VisualizerUpload
        
        """
        f"Default configuration file is at {config.DEFAULT_CONFIG_FILE}"
    )
    ap.add_argument('-c', type=str, help='Use as alternate config file')

    args = ap.parse_args()

    pyde1_logging.setup_initial_logger()

    config.load_from_yaml(args.c)

    pyde1_logging.setup_direct_logging(config.logging)
    pyde1_logging.config_logger_levels(config.logging)

    logger = pyDE1.getLogger('Main')

    logger_mqtt = pyDE1.getLogger('MQTT')
    logger_upload = pyDE1.getLogger('Upload')

    loop = asyncio.get_event_loop()
    loop.set_debug(True)
    loop.set_exception_handler(sm.exception_handler)
    sm.attach_signal_handler_to_loop(sm.shutdown, loop)

    t = loop.create_task(setup_and_run())

    loop.run_forever()
    exit(sm.exit_value)





"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only
"""

import asyncio
import copy
import inspect
import math  # for nan to be uniquely math.nan
import time
from functools import reduce
from typing import Union, Dict, Set

import pyDE1.scanner
from pyDE1.config import config
from pyDE1.de1 import DE1
from pyDE1.de1.c_api import PackedAttr, MMR0x80LowAddr, pack_one_mmr0x80_write
from pyDE1.de1.notifications import MMR0x80Data
from pyDE1.dispatcher.mapping import MAPPING, TO, IsAt
from pyDE1.dispatcher.resource import Resource
from pyDE1.exceptions import (
    DE1APITypeError, DE1APIValueError, DE1APIAttributeError, DE1APIKeyError,
    DE1NotConnectedError
)
from pyDE1.flow_sequencer import FlowSequencer
from pyDE1.scale.processor import ScaleProcessor
from pyDE1.scanner import DiscoveredDevices, scan_from_api
from pyDE1.utils import prep_for_json
from pyDE1.utils_public import rgetattr, rsetattr

logger = pyDE1.getLogger('Inbound.Implementation')

"""
How this works:

http/run.py creates an APIRequest and adds it to the queue

The _request_queue_processor from dispatcher picks it up and calls
GET:        resource_dict = await get_resource_to_dict(got.resource)
PATCH, PUT: await patch_resource_from_dict(got.resource, got.payload)

The resource_dict becomes a payload for the APIResponse object that is queued

To be able to return meaningful results from PUT/PATCH, the recipient 
needs to know which of possibly many elements the result comes from.
As most won't be returning anything, at least right now, append each,
within a nested dict, to a list and pass that back.
"""


def get_timeout(prop, value):
    bound_class = prop.__self__.__class__
    name = prop.__name__
    timeout = config.http.ASYNC_TIMEOUT
    if name == 'connectivity_setter' and value:
        timeout = config.bluetooth.CONNECT_TIMEOUT + 0.100
    elif name in ('change_de1_to_id', 'change_scale_to_id'):
        if value is None:
            timeout = config.bluetooth.DISCONNECT_TIMEOUT
        else:
            timeout = config.bluetooth.CONNECT_TIMEOUT \
                      + config.http.ASYNC_TIMEOUT + 0.100
    elif name == 'first_if_found':
        timeout = config.bluetooth.SCAN_TIME \
                  + config.bluetooth.CONNECT_TIMEOUT \
                  + config.http.ASYNC_TIMEOUT + 0.100
    elif name == 'upload_json_v2_profile':
        timeout = config.http.PROFILE_TIMEOUT
    elif name == 'stop_at_time_set_async':
        timeout = config.http.ASYNC_TIMEOUT * 2

    elif name == 'upload_firmware_from_content':
        timeout = config.http.FIRMWARE_TIMEOUT

    if DE1().uploading_firmware:
        timeout += config.de1.CUUID_LOCK_WAIT_TIMEOUT

    return timeout


async def _prop_value_setter(prop, value):

    if inspect.isroutine(prop):
        if inspect.iscoroutinefunction(prop):
            retval = await asyncio.wait_for(prop(value),
                                            get_timeout(prop, value))
        else:
            retval = prop(value)
    else:
        raise DE1APITypeError (f"Setter {prop} is not callable")

    return retval


async def _prop_value_getter(prop):

    if inspect.isroutine(prop):
        if inspect.iscoroutinefunction(prop):
            retval = await asyncio.wait_for(prop(),
                                            get_timeout(prop, None))
        else:
            retval = prop()
    else:
        raise DE1APITypeError (f"Setter {prop} is not callable")

    return retval


# NB: This assumes that the MMR and CUUID are kept up to date
#     and that those that are read don't change on their own

async def _get_isat_value(isat: IsAt):

    # TODO: if IsAt.use_getter is implemented in the future, change here

    target = isat.target
    attr_path = isat.attr_path
    if attr_path is None:
        raise DE1APIAttributeError(f"Write-only attribute {isat.__repr__()}")

    de1 = DE1()
    flow_sequencer = FlowSequencer()
    scale_processor = ScaleProcessor()
    scale = scale_processor.scale
    dd = DiscoveredDevices()

    retval = None

    # For any attribute or property with a getter, getattr() "just works"
    # If possibly a callable, need to
    #     await _prop_value_getter(rgetattr(target, attr_path))

    if target == TO.DE1:
        retval = rgetattr(de1, attr_path)

    elif target == TO.FlowSequencer:
        retval = rgetattr(flow_sequencer, attr_path)

    elif target == TO.Scale:
        retval = rgetattr(scale, attr_path)

    elif target == TO.ScaleProcessor:
        retval = rgetattr(scale_processor, attr_path)

    elif target == TO.DiscoveredDevices:
        # Need to wrap as devices_for_json is async
        retval = await _prop_value_getter(rgetattr(dd, attr_path))

    elif isinstance(target, MMR0x80LowAddr):
        # NB: This assumes that the MMR and CUUID are kept up to date
        #     and that those that are read don't change on their own

        if not de1.is_ready:
            raise DE1NotConnectedError(
                "DE1 is not connected at last-chance check")

        if attr_path != '':
            raise DE1APIAttributeError(
                "MMR reads do not support attr_path "
                f"{isat}")

        # For now, assume everything is kept current

        # TODO: Unify with replication in PATCH operation

        if target.value > de1.feature_flag.last_mmr0x80:
            retval = None
            logger.info(
                f"Skipping (not in FW) {target.name}, "
                f"0x{target.value:04x} > 0x{de1.feature_flag.last_mmr0x80:04x}")
        else:
            try:
                retval = de1._mmr_dict[target].data_decoded
            except KeyError:
                retval = None
            if retval is None:
                t0 = time.time()
                # TODO: Can this be simplified/clarified?
                ready = await de1.read_one_mmr0x80(target)
                await ready.wait()
                retval = de1._mmr_dict[target].data_decoded
                t1 = time.time()
                logger.debug(
                    f"Read of {target.__repr__()} took \t"
                    f"{(t1 - t0) * 1000:6.1f} ms"
                )

    elif inspect.isclass(target) and issubclass(target, PackedAttr):
        # NB: This assumes that the MMR and CUUID are kept up to date
        #     and that those that are read don't change on their own

        if not de1.is_ready:
            raise DE1NotConnectedError(
                "DE1 is not connected at last-chance check")

        try:
            obj = (de1._cuuid_dict[target.cuuid]).last_value
        except KeyError:
            logger.info(
                f"No last value for {target.cuuid}"
            )
            obj = None
        if obj is None:
            t0 = time.time()
            obj = await de1.read_cuuid(target.cuuid)
            t1 = time.time()
            logger.debug(
                f"Read of {target} took \t{(t1 - t0) * 1000:6.1f} ms"
            )
        retval = rgetattr(obj, attr_path)

    else:
        raise DE1APITypeError(
            f"Mapping target of {target} is not recognized"
        )

    return prep_for_json(retval)


async def _get_mapping_to_dict(partial_dict: dict) -> dict:
    """
    Takes a "branch" of a mapping and
      * Fills in any IsAt values
      * Recursively calls itself if a dict
      * Passes any other values unmodified
    """

    if isinstance(partial_dict, IsAt):
        partial_dict = { None: partial_dict }

    if not isinstance(partial_dict, dict):
        raise DE1APITypeError(f"Expected a dict, not {type(partial_dict)}")

    retval = {}

    for k, v in partial_dict.items():
        if isinstance(v, IsAt):
            try:
                this_val = await _get_isat_value(v)
            except AttributeError:
                if config.http.PRUNE_EMPTY_NODES:
                    continue # Don't write the key's entry
                else:
                    this_val = math.nan
        elif isinstance(v, dict):
            this_val = await _get_mapping_to_dict(v)
            # Suppress aggregates with nothing to aggregate
            if len(this_val) == 0 and config.http.PRUNE_EMPTY_NODES:
                continue
        else:
            this_val = v
        retval[k] = this_val

    return retval

# TODO: set and get versions (no recollection of what this means)

def _get_target_sets_inner(partial_dict: dict,
                           dict_of_sets: Dict[
                               str, Set[Union[PackedAttr, MMR0x80LowAddr]]],
                           include_can_read: bool,
                           include_can_write: bool):
    """
    Takes a "branch" of a mapping and returns a dict with two keys,
    'PacketAttr' and 'MMR0x80LowAddr', each with a set of targets.
    The dict_of_sets is modified in-place.
    """

    # See also validate.py
    # Valid: dict with dict
    #        IsAt with byte, bytearray (profile or firmware)
    if isinstance(partial_dict, IsAt):
        partial_dict = { None: partial_dict }

    if not isinstance(partial_dict, dict):
        raise DE1APITypeError(f"Expected a dict, not {type(partial_dict)}")

    for k, isat in partial_dict.items():

        if isinstance(isat, IsAt):
            target = isat.target
            if not ((inspect.isclass(target) and issubclass(target, PackedAttr))
                    or isinstance(target, MMR0x80LowAddr)):
                continue
            writable = target.can_write \
                       and not (isat.read_only
                                or (isat.attr_path is None
                                    and isat.setter_path is None))
            readable = target.can_read and isat.attr_path is not None
            if ((include_can_read and readable)
                    or (include_can_write and writable)):
                if isinstance(target, MMR0x80LowAddr):
                    dict_of_sets['MMR0x80LowAddr'].add(target)
                elif inspect.isclass(target) and issubclass(target, PackedAttr):
                    dict_of_sets['PackedAttr'].add(target)
            else:
                if not readable and not writable:
                    logger.error(
                        "IsAt should have either or both "
                        f"attr_path or setter_path not None: {isat.__repr__()}"
                    )

        elif isinstance(isat, dict):
            _get_target_sets_inner(isat, dict_of_sets,
                                   include_can_read, include_can_write)


def get_target_sets(mapping: dict,
                    include_can_read=False, include_can_write=False) \
        -> Dict[str, Set[Union[MMR0x80LowAddr, PackedAttr]]]:

    retval = {
        'MMR0x80LowAddr': set(),
        'PackedAttr': set()
    }

    _get_target_sets_inner(mapping, retval,
                           include_can_read, include_can_write)
    return retval


async def get_resource_to_dict(resource: Resource) -> dict:

    mapping = MAPPING[resource]
    return await _get_mapping_to_dict(mapping)


# PATCH and PUT are related, but have slightly different requirements
#
# For PATCH, it is sufficient that each entry from the request has
# a corresponding entry in the overall mapping dictionary
#
# For PUT it is required that each element in the mappig dictionary
# have a corresponding entry in the request dictionary (completeness)
#

# TODO: How to handle read-only attributes in the PUT case is TBD

async def patch_resource_from_dict(resource: Resource, values_dict: dict):

    mapping = MAPPING[resource]

    # Get the list of PackedAttrs that need to be patched (properties and
    # MMRs are handled one at a time, as they are atomic), lock from changes,
    # get copies of current PackedAttrs to patch

    # TODO: Managing locks
    # TODO: PackedAttr "getter" that will wait on a pending update, if one is
    #       in flight, as well as unify the retrieval if not present

    # target_sets = get_target_sets(values_dict, include_can_write=True)

    # TODO: Optimize this -- this needs a mapping, the way it is written now
    #       It really should only retrieve those that are being changed
    #       in the case of a PATCH

    target_sets = get_target_sets(mapping, include_can_write=True)

    # Lock here
    de1 = DE1()
    flow_sequencer = FlowSequencer()

    for pa in target_sets['PackedAttr']:
        pa: PackedAttr
        cuuid = pa.cuuid
        try:
            last_value = de1._cuuid_dict[cuuid]._last_value
        except KeyError:
            last_value = None
        if last_value is None:
            t0 = time.time()
            last_value = await de1.read_cuuid(cuuid)
            t1 = time.time()
            logger.debug(
                f"Read of {cuuid} took \t"
                f"{(t1 - t0) * 1000:6.1f} ms"
            )

        # Don't load them all, just make sure they are there
        # Only load if they are being changed
        # pending_packed_attrs[pa] = last_value

    pending_packed_attrs = {}

    # Valid: dict with dict
    #        IsAt with byte, bytearray (profile or firmware)
    if isinstance(mapping, dict) \
            and isinstance(values_dict, dict):
        pass
    elif isinstance(mapping, IsAt) \
        and isinstance(values_dict, (bytes, bytearray)):
        # coerce into "standard form"
        mapping = { None: mapping }
        values_dict = { None: values_dict }
    else:
        raise DE1APITypeError(
            "Mapping and patch inconsistent, "
            "dict with dict, IsAt with raw value "
            f"not {type(mapping)} with {type(values_dict)}"
        )

    results_list = list()
    await _patch_dict_to_mapping_inner(values_dict,
                                       mapping,
                                       pending_packed_attrs,
                                       list(),
                                       results_list)

    # if there are pending_packed_attrs, send them

    # Potentially gather, but may not be faster
    for pa in pending_packed_attrs.values():
        await de1.write_packed_attr(pa)

    # release locks

    return results_list


async def _patch_dict_to_mapping_inner(partial_value_dict: dict,
                                       partial_mapping_dict: dict,
                                       pending_packed_attrs: Dict[
                                           type(PackedAttr), PackedAttr],
                                       running_path: list,
                                       results_list: list):

    """
    This assumes that everything has been determined as "valid"
    """

    # TODO: PATCH Method for HTTP specifies that PATCH is to be atomic
    #       This isn't completely possible as there is no
    #       roll-back capability implemented at this time.
    #       To at least be somewhat compliant, all keys
    #       should be processed as fully as possible
    #       prior to writing any to their targets.
    #
    #       https://datatracker.ietf.org/doc/html/rfc5789
    #
    #       NB: Although this is similar to JSON Merge Patch
    #       https://datatracker.ietf.org/doc/html/rfc7386
    #       it is subtly different in that a JSON 'null'
    #       is interpreted as the valid, Python 'None'
    #       rather than as a request to delete an element:
    #
    #           This design means that merge patch documents are suitable for
    #           describing modifications to JSON documents that primarily
    #           use objects for their structure and do not make use of
    #           explicit null values.
    #
    #           The merge patch format is not appropriate for all JSON syntaxes.

    de1 = DE1()
    flow_sequencer = FlowSequencer()
    scale_processor = ScaleProcessor()
    scale = scale_processor.scale

    for key, new_value in partial_value_dict.items():

        try:
            mapping_isat = partial_mapping_dict[key]
        except KeyError:
            raise DE1APIKeyError(
                f"Unable to find mapping for {key} on the specified path"
            )

        if isinstance(mapping_isat, IsAt):
            target = mapping_isat.target
            attr_path = mapping_isat.attr_path
            setter_path = mapping_isat.setter_path

            if mapping_isat.read_only or attr_path is None and setter_path is None:
                raise DE1APIValueError(
                    f"Mapping for '{key}': {mapping_isat} is not writable"
                )

            if isinstance(target, TO) and target != TO.DiscoveredDevices:
                if target == TO.DE1:
                    this_target = de1
                elif target == TO.FlowSequencer:
                    this_target = flow_sequencer
                elif target == TO.Scale:
                    this_target = scale
                elif target == TO.ScaleProcessor:
                    this_target = scale_processor
                else:
                    raise DE1APITypeError(
                        f"Unsupported target for '{key}': {mapping_isat}")

                if setter_path is not None:
                    # Allow for a non-property setter to return a value
                    setter = rgetattr(this_target, setter_path)
                    retval = await _prop_value_setter(setter, new_value)
                    if retval is not None:
                        # reduce(lambda a, b: {b: a}, [5,4,3,2,1], 'val')
                        #     {1: {2: {3: {4: {5: 'val'}}}}}
                        result_dict = reduce(lambda a, b: {b: a},
                                             running_path,
                                             retval)
                        results_list.append(result_dict)

                else:
                    rsetattr(this_target, attr_path, new_value)

            # TODO: Is there a better way to work with an unbound function?
            #       Maybe attach it to a module, rathern than a special case?
            elif target is None:
                if setter_path == 'scan_from_api':
                    retval = await scan_from_api(new_value)
                    if retval is not None:
                        # reduce(lambda a, b: {b: a}, [5,4,3,2,1], 'val')
                        #     {1: {2: {3: {4: {5: 'val'}}}}}
                        result_dict = reduce(lambda a, b: {b: a},
                                             running_path,
                                             retval)
                        results_list.append(result_dict)

            elif isinstance(target, MMR0x80LowAddr):

                # MMR writes need to be serial and are atomic in that
                # each writable MMR is a single value.
                # As a result, just write it here and now.

                if attr_path != '':
                    raise DE1APIValueError(
                        "MMR writes do not support attr_path "
                        f"{target} {mapping_isat}")

                if setter_path is not None:
                    raise DE1APIValueError(
                        "MMR writes do not support setter_path "
                        f"{target} {mapping_isat}")

                mmr_write = pack_one_mmr0x80_write(
                    addr_low= target,
                    value= new_value,
                )

                logger.debug(f"MMR to be written: {mmr_write.as_wire_bytes()}")

                await de1.write_packed_attr(mmr_write)

                # TODO: Should this wait on ready.wait() ??
                #       Or is there a way to collect them all for later
                #       as a speed optimization?
                ns: MMR0x80Data = de1._mmr_dict[target]
                await ns.ready_event.wait()

            elif inspect.isclass(target) and issubclass(target, PackedAttr):
                # NB: This assumes that the CUUIDs are kept up to date
                #     and that those that are read don't change on their own

                packed_attr = (de1._cuuid_dict[target.cuuid]).last_value
                # TODO: Can this be sped up reliably?
                if packed_attr is None:
                    packed_attr = await de1.read_cuuid(target.cuuid)
                old_value = rgetattr(packed_attr, attr_path)
                if new_value != old_value:
                    if not target in pending_packed_attrs:
                        pending_packed_attrs[target] = copy.deepcopy(packed_attr)
                    rsetattr(pending_packed_attrs[target], attr_path, new_value)

                # Send in outer method once all nodes are visited

            else:
                raise DE1APITypeError(
                    f"Mapping target of {target} is not recognized")

        elif isinstance(mapping_isat, dict):
            # TODO: Where did "this_val" (unused) come from?
            this_val = await _patch_dict_to_mapping_inner(
                partial_value_dict[key],
                partial_mapping_dict[key],
                pending_packed_attrs,
                # Prepend as reduce needs "reversed" list and need copy anyway
                running_path=[mapping_isat] + running_path,
                results_list=results_list)
        else:
            this_val = new_value



"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only
"""

import asyncio
import enum
import logging
import re
import uuid

from datetime import datetime

logger_cancel_tasks = logging.getLogger('Util.CancelTasks')


def task_name_exists(name_to_find: str, starts_with=False):
    all_tasks = asyncio.all_tasks()
    found = False
    for t in all_tasks:
        name = t.get_name()
        if starts_with:
            found = name.startswith(name_to_find)
        else:
            found = name == name_to_find
        if found:
            break
    return found


def cancel_tasks_by_name(name_to_cancel: str, starts_with=False):
    all_tasks = asyncio.all_tasks()
    me = asyncio.current_task()
    for t in all_tasks:
        if t is me:
            continue
        name = t.get_name()
        if starts_with:
            cancel = name.startswith(name_to_cancel)
        else:
            cancel = name == name_to_cancel
        if cancel:
            logger_cancel_tasks.info(f"Canceling {t}")
            t.cancel()


def address_is_persistent(address: str) -> bool:
    """
    CoreBluetooth uses a UUID rather than a MAC address to identify devices
    As a result, it is (probably) not persistent over reboots.
    """
    if re.match(r'([0-9a-fA-F]{2}:){5}[0-9a-fA-F]{2}$', address):
        return True
    else:
        # macOS UUIDs don't seem to be persistent across Mac reboots
        return False


def prep_for_json(val):
    """
    Special cases for conversion to JSON:
    * enum classes
    * bytes-like
    * UUID

    Return the name of an IntEnum, does not help with IntFlag
    So far no IntFlag enums headed to the external API
    """
    # Order is important due to IntEnum and IntFlag behavior (including int)
    if val is None or isinstance(val, (float, str, bool)):
        return val
    elif isinstance(val, enum.IntFlag):
        # Remove the class name
        return str(val).split('.',2)[1]
    elif isinstance(val, enum.IntEnum):
        return val.name
    elif isinstance(val, enum.Enum):
        return val.value
    elif isinstance(val, (bytearray, bytes)):
        return val.hex()
    elif isinstance(val, uuid.UUID):
        return str(val)
    else:
        return val


def data_as_hex(data):
    hex_data = data.hex()
    return ' '.join(b0 + b1 for b0, b1
                    in zip(hex_data[0::2], hex_data[1::2]))


re_data_is_ascii_readable = re.compile('^[\x20-\x7e]*$')
re_data_is_ascii_readable_with_subs = re.compile('^[\r\n\t\x20-\x7e]*$')
tt_rnt_glyphs = str.maketrans("\r\n\t", "\u240d\u240a\u2409")
tt_space_glpyh = str.maketrans(" ", "\u2423")


def data_as_readable(data, replace_rnt=True, replace_space=False):
    """
    Convert a string, bytes, or bytearray object to a "readable" string
    if able to do so, return '' if not readable.

    "readable" is only containing the ASCII characters [space] through ~ (tilde)
    potentially with substitution of glyphs for \r \n \t and/or [space]

    Ref: http://www.unicode.org/charts/PDF/U2400.pdf

    :param data:            str, bytes, or bytearray to be converted
    :param replace_rnt:     Allow \r \n \t and substitute ␍ ␊ ␉
                            (considered not readable otherwise)
    :param replace_space:   Substitute ␣ for [space]
    :return:                str: converted to readable, or ''
    """
    if isinstance(data, (bytes, bytearray)):
        try:
            data = data.decode('ascii')  # enforce one character per byte
        except UnicodeDecodeError:
            return ''
    if not isinstance(data, str):
        raise TypeError("Expected str, bytes, or bytearray")
    out = ""
    if replace_rnt:
        if re_data_is_ascii_readable_with_subs.match(data):
            out = data.translate(tt_rnt_glyphs)
    else:
        if re_data_is_ascii_readable.match(data):
            out = data
    if replace_space:
        out = out.translate(tt_space_glpyh)
    return out


def data_as_readable_or_hex(data, replace_rnt=True, replace_space=False):
    """
    Convert a string, bytes, or bytearray object to a "readable" string
    if able to do so, return data_as_hex(data) if not readable.

    "readable" is only containing the ASCII characters [space] through ~ (tilde)
    potentially with substitution of glyphs for \r \n \t and/or [space]

    Ref: http://www.unicode.org/charts/PDF/U2400.pdf

    :param data:            str, bytes, or bytearray to be converted
    :param replace_rnt:     Allow \r \n \t and substitute ␍ ␊ ␉
                            (considered not readable otherwise)
    :param replace_space:   Substitute ␣ for [space]
    :return:                str: converted to readable, else data_as_hex(data)
    """
    if len(data) == 0:
        return ''
    out = data_as_readable(data, replace_rnt, replace_space)
    if len(out) == 0:
        if not isinstance(data, (bytes, bytearray)):
            # Try to be helpful and assume it is byte-by-byte
            # data_as_readable() has already confirmed str, bytes, bytearray
            data = data.encode('ascii')
        out = data_as_hex(data)
    return out


def timestamp_to_str_with_ms(timestamp: float, show_date=True) -> str:
    string = datetime.fromtimestamp(
        timestamp).isoformat(sep=' ', timespec='milliseconds')
    if not show_date:
        (d, t) = string.split(' ', maxsplit=2)
        string = t
    return string





"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only

Replays the captured packets from a flow sequence
such as for testing or demonstration of "GUIs"

NB: Does _not_ modify the DB contents, so real-time pulls by the consumer
    will be "at completion", not "as they were".

    This primarily impacts the sequence table, which is updated
    as the sequence unfolds.
"""

import json
import logging
import os
import queue
import socket
import sqlite3
import time
import asyncio

from typing import NamedTuple, Union, List, Optional

import paho.mqtt.client as mqtt
from paho.mqtt.client import MQTTv5, MQTT_CLEAN_START_FIRST_ONLY

import pyDE1
import pyDE1.pyde1_logging as pyde1_logging
from pyDE1.config_load import ConfigYAML, ConfigLoadable
from pyDE1.pyde1_logging import ConfigLogging, ConfigLoggingFormatters, \
    ConfigLoggingHandlers


class Config (ConfigYAML):

    DEFAULT_CONFIG_FILE = '/usr/local/etc/pyde1/pyde1-replay.conf'

    def __init__(self):
        super(Config, self).__init__()
        self.database = _Database()
        self.logging = _Logging()
        self.mqtt = _MQTT()
        self.sequence = _Sequence()


# This craziness is so pyCharm autocompletes
# Otherwise typing.SimpleNamespace() would be sufficient

class _MQTT (ConfigLoadable):
    def __init__(self):
        self.TOPIC_ROOT = 'KEpyDE1'
        self.CLIENT_ID_PREFIX = 'pyde1-replay'
        self.BROKER_HOSTNAME = '::1'
        self.BROKER_PORT = 1883
        self.TRANSPORT = 'tcp'
        self.TLS_CONTEXT = None
        self.KEEPALIVE = 60
        self.USERNAME = None
        self.PASSWORD = None
        self.DEBUG = False


class _Logging (ConfigLogging):
    def __init__(self):
        super(_Logging, self).__init__()
        # NB: The log file name is matched against [a-zA-Z0-9._-]
        self.LOG_FILENAME = 'replay.log'
        self.LOGGERS = {
        }

# Accept default formatters and handlers


class _Database (ConfigLoadable):
    def __init__(self):
        self.FILENAME = '/var/lib/pyde1/pyde1.sqlite3'


class _Sequence (ConfigLoadable):
    def __init__(self):
        self.ID = None


config = Config()


# TODO: Figure out how not to duplicate this in so many places

# NB: Remember to reconstruct the class element

class SequenceRow (NamedTuple):
    id: str
    active_state:   str
    start_sequence: float
    start_flow:     float
    end_flow:       float
    end_sequence:   float
    profile_id:     str
    # https://www.sqlite.org/quirks.html#no_separate_boolean_datatype
    profile_assumed:    int     # 0: False, 1: True
    resource_version:                           str
    resource_de1_id:                            str
    resource_de1_read_once:                     str
    resource_de1_calibration_flow_multiplier:   str
    resource_de1_control_mode:                  str
    resource_de1_control_tank_water_threshold:  str
    resource_de1_setting_before_flow:           str
    resource_de1_setting_steam:                 str
    resource_de1_setting_target_group_temp:     str
    resource_scale_id:                          str

    @property
    def class_str(self):
        return None


def sequence_row_factory(cur: sqlite3.Cursor, row: sqlite3.Row):
    return SequenceRow(*row)


class ShotSampleRow (NamedTuple):
    version:            str
    sender:             str
    arrival_time:       float
    create_time:        float
    event_time:         float
    #
    de1_time:           float
    sample_time:        int
    group_pressure:     float
    group_flow:         float
    mix_temp:           float
    head_temp:          float
    set_mix_temp:       float
    set_head_temp:      float
    set_group_pressure: float
    set_group_flow:     float
    frame_number:       int
    steam_temp:         float
    volume_preinfuse:   float
    volume_pour:        float
    volume_total:       float
    volume_by_frames:   str     # representation of list

    @property
    def class_str(self):
        return 'ShotSampleWithVolumesUpdate'


def shot_sample_row_factory(cur: sqlite3.Cursor, row: sqlite3.Row):
    return ShotSampleRow(*row)


class WeightFlowRow (NamedTuple):
    version:            str
    sender:             str
    arrival_time:       float
    create_time:        float
    event_time:         float
    #
    scale_time:             float
    current_weight:         float
    current_weight_time:    float
    average_flow:           float
    average_flow_time:      float
    median_weight:          float
    median_weight_time:     float
    median_flow:            float
    median_flow_time:       float

    @property
    def class_str(self):
        return 'WeightAndFlowUpdate'


def weight_flow_row_factory(cur: sqlite3.Cursor, row: sqlite3.Row):
    return WeightFlowRow(*row)


class StateUpdateRow (NamedTuple):
    version:            str
    sender:             str
    arrival_time:       float
    create_time:        float
    event_time:         float
    #
    event_time:         float
    state:              str
    substate:           str
    previous_state:     str
    previous_substate:  str
    is_error_state:     str     # TODO: Change to bool in schema and access
                                #       2021-10 -- Would require update trigger
                                #       Consider at next schema change

    @property
    def class_str(self):
        return 'StateUpdate'


def state_update_row_factory(cur: sqlite3.Cursor, row: sqlite3.Row):
    return StateUpdateRow(*row)


class SequencerGateNotificationRow (NamedTuple):
    version:            str
    sender:             str
    arrival_time:       float
    create_time:        float
    event_time:         float
    #
    name:           str
    action:         str
    active_state:   str
    sequence_id:    str

    @property
    def class_str(self):
        return 'SequencerGateNotification'


def sequence_gate_notification_row_factory(cur: sqlite3.Cursor,
                                           row: sqlite3.Row):
    return SequencerGateNotificationRow(*row)


class WaterLevelRow (NamedTuple):
    version:            str
    sender:             str
    arrival_time:       float
    create_time:        float
    event_time:         float
    #
    level:              float
    start_fill_level:   float

    @property
    def class_str(self):
        return 'WaterLevelUpdate'


def water_level_row_factory(cur: sqlite3.Cursor, row: sqlite3.Row):
    return WaterLevelRow(*row)


# NB: Database does not yet not include name or ID

class OldConnectivityChangeRow (NamedTuple):
    version:            str
    sender:             str
    arrival_time:       float
    create_time:        float
    event_time:         float
    #
    state:              str
    # name:               str
    # id:                 str


def old_connectivity_change_row_factory(cur: sqlite3.Cursor,
                                        row: sqlite3.Row):
    return OldConnectivityChangeRow(*row)


class ConnectivityChangeRow (NamedTuple):
    version:            str
    sender:             str
    arrival_time:       float
    create_time:        float
    event_time:         float
    #
    state:              str
    name:               str
    id:                 str

    @property
    def class_str(self):
        return 'ConnectivityChange'


def augment_old_connectivity_change_row(
        old_row: OldConnectivityChangeRow) -> ConnectivityChangeRow:

    if old_row.sender == 'DE1':
        id = 'D9:B2:48:aa:bb:cc'
        name = 'DE1'
    else:
        id = 'CF:75:75:aa:bb::cc'
        name = 'Skale'

    return ConnectivityChangeRow(*old_row, name, id)


class SendListEntry (NamedTuple):
    send_at:    float
    payload:    str


def _shift_if_time(key: str, val: Union[str, float], shift: float):
    if key.endswith('_time'):
        return val + shift
    else:
        return val


def create_entry(row: NamedTuple, shift_time: float) -> SendListEntry:
    """
    Shift all time elements by adding shift_time
    Return a JSON string, so compatible with api/mqtt/run
        outbound_pipe_reader()
            item_json = outbound_pipe.recv()
    """

    row_dict = {k:_shift_if_time(k, v, shift_time)
                for (k,v) in row._asdict().items()}
    # For now, bomb out on missing property
    row_dict['class'] = row.class_str
    row_dict['shifted'] = shift_time

    return SendListEntry(send_at=row_dict['event_time'],
                         payload=json.dumps(row_dict))


def collect_send_list(sequence_id: str,
                      shift_time: float) -> List[SendListEntry]:
    send_list = []

    with sqlite3.connect(f"file:{config.database.FILENAME}?mode=ro",
                         uri=True) as db:

        db.row_factory = shot_sample_row_factory
        cur = db.execute(f"SELECT {' ,'.join(ShotSampleRow._fields)} "
                         "FROM shot_sample_with_volume_update "
                         "WHERE sequence_id == :id "
                         "ORDER BY event_time",
                         {'id': sequence_id})
        for row in cur.fetchall():
            send_list.append(create_entry(row, shift_time))

        db.row_factory = weight_flow_row_factory
        cur = db.execute(f"SELECT {' ,'.join(WeightFlowRow._fields)} "
                         "FROM weight_and_flow_update "
                         "WHERE sequence_id == :id "
                         "ORDER BY event_time",
                         {'id': sequence_id})
        for row in cur.fetchall():
            send_list.append(create_entry(row, shift_time))

        db.row_factory = state_update_row_factory
        cur = db.execute(f"SELECT {' ,'.join(StateUpdateRow._fields)} "
                         "FROM state_update "
                         "WHERE sequence_id == :id "
                         "ORDER BY event_time",
                         {'id': sequence_id})
        for row in cur.fetchall():
            send_list.append(create_entry(row, shift_time))

        db.row_factory = water_level_row_factory
        cur = db.execute(f"SELECT {' ,'.join(WaterLevelRow._fields)} "
                         "FROM water_level_update "
                         "WHERE sequence_id == :id "
                         "ORDER BY event_time",
                         {'id': sequence_id})
        for row in cur.fetchall():
            send_list.append(create_entry(row, shift_time))

        db.row_factory = old_connectivity_change_row_factory
        cur = db.execute(
            f"SELECT {' ,'.join(OldConnectivityChangeRow._fields)} "
            "FROM connectivity_change "
            "WHERE sequence_id == :id "
            "ORDER BY event_time",
            {'id': sequence_id})
        for row in cur.fetchall():
            new_row = augment_old_connectivity_change_row(row)
            send_list.append(create_entry(new_row, shift_time))

        db.row_factory = sequence_gate_notification_row_factory
        cur = db.execute(
            f"SELECT {' ,'.join(SequencerGateNotificationRow._fields)} "
            "FROM sequencer_gate_notification "
            "WHERE sequence_id == :id "
            "ORDER BY event_time",
            {'id': sequence_id})
        for row in cur.fetchall():
            send_list.append(create_entry(row, shift_time))

    send_list.sort(key=lambda entry: entry.send_at)
    return send_list


def get_sequence_start_time(sequence_id: str) -> float:

    with sqlite3.connect(f"file:{config.database.FILENAME}?mode=ro",
                         uri=True) as db:
        db.row_factory = sequence_row_factory
        cur = db.execute(f"SELECT {', '.join(SequenceRow._fields)} "
                         "FROM sequence "
                         "WHERE id == :id",
                         (sequence_id,))
        row = cur.fetchone()
        return row.start_sequence


# MQTT

def setup_client(mqtt_client_logger: logging.Logger) -> mqtt.Client:

    def on_log_callback(client: mqtt.Client, userdata, level, buf):
        mqtt_client_logger.info(f"CB: Log: level: {level} '{buf}' ({type(buf)})")

    def on_connect_callback(client, userdata, flags, reasonCode, properties):
        mqtt_client_logger.info(
            f"CB: Connect: flags: {flags}, reasonCode: {reasonCode}, "
            f"properties {properties}")

    def on_publish_callback(client, userdata, mid):
        mqtt_client_logger.info(f"CB: Published: mid: {mid}")

    # Caught exception in on_disconnect:
    #     on_disconnect_callback() missing 1 required positional argument:
    #         'properties'
    def on_disconnect_callback(client, userdata, reasonCode, properties=None):
        mqtt_client_logger.info(f"CB: Disconnect: reasonCode: {reasonCode}, "
                           f"properties {properties}")

    def on_socket_open_callback(client, userdata, socket):
        mqtt_client_logger.info(f"CB: Socket open: socket: {socket}")

    def on_socket_close_callback(client, userdata, socket):
        mqtt_client_logger.info(f"CB: Socket close: socket: {socket}")

    def on_socket_register_write_callback(client, userdata, socket):
        mqtt_client_logger.info(f"CB: Socket register write: socket: {socket}")

    def on_socket_unregister_write_callback(client, userdata, socket):
        mqtt_client_logger.info(f"CB: Socket unregister write: socket: {socket}")

    mqtt_client = mqtt.Client(
        client_id="{}@{}[{}]".format(
            config.mqtt.CLIENT_ID_PREFIX,
            socket.gethostname(),
            os.getpid(),
        ),
        clean_session=None,  # Required for MQTT5
        userdata=None,
        protocol=MQTTv5,
        transport=config.mqtt.TRANSPORT,
    )

    if config.mqtt.USERNAME is not None:
        mqtt_client_logger.info(
            f"Connecting with username '{config.mqtt.USERNAME}'")
        mqtt_client.username_pw_set(
            username=config.mqtt.USERNAME,
            password=config.mqtt.PASSWORD
        )

    # mqtt_client.on_log = on_log_callback
    mqtt_client.on_connect = on_connect_callback
    # mqtt_client.on_publish = on_publish_callback
    mqtt_client.on_disconnect = on_disconnect_callback
    mqtt_client.on_socket_open = on_socket_open_callback
    mqtt_client.on_socket_close = on_socket_close_callback
    # mqtt_client.on_socket_register_write = on_socket_register_write_callback
    # mqtt_client.on_socket_unregister_write = on_socket_unregister_write_callback

    mqtt_client.enable_logger(mqtt_client_logger)

    result = mqtt_client.connect(host=config.mqtt.BROKER_HOSTNAME,
                   port=config.mqtt.BROKER_PORT,
                   keepalive=config.mqtt.KEEPALIVE,
                   bind_address="",
                   bind_port=0,
                   clean_start=MQTT_CLEAN_START_FIRST_ONLY,
                   properties=None)
    client_logger.info(f"connect returned {result}")

    return mqtt_client


if __name__ == '__main__':

    import argparse

    ap = argparse.ArgumentParser(
        description=
        """Replay a sequence from the database, time shifted to the present.
        
        """
        f"Default configuration file is at {config.DEFAULT_CONFIG_FILE}"
    )
    ap.add_argument('-c', type=str, help='Use as alternate config file')
    ap.add_argument('-s', type=str, help='Override for sequence ID')
    ap.add_argument('-t', type=str, help='Override for MQTT topic')

    args = ap.parse_args()

    pyde1_logging.setup_initial_logger()

    config.load_from_yaml(args.c)

    if args.s is not None:
        config.sequence.ID = args.s

    if args.t is not None:
        config.mqtt.TOPIC_ROOT = args.t

    pyde1_logging.setup_direct_logging(config.logging)
    pyde1_logging.config_logger_levels(config.logging)

    logger = pyDE1.getLogger('Replay')
    client_logger = pyDE1.getLogger('MQTTClient')

    sst = get_sequence_start_time(config.sequence.ID)
    now = time.time()
    start_sequence_at = now + 5
    shift_time = start_sequence_at - sst
    send_list = collect_send_list(config.sequence.ID, shift_time)
    mqtt_client = setup_client(client_logger)
    mqtt_client.loop_start()

    MQTT_LEAD_TIME = 0.000  # seconds

    result = None

    while len(send_list):
        next_to_send = send_list.pop(0)
        while next_to_send.send_at > time.time() + MQTT_LEAD_TIME:
            time.sleep(0.010)
        # print(time.time(), next_to_send)
        item_as_dict = json.loads(next_to_send.payload)
        topic = f"{config.mqtt.TOPIC_ROOT}/{item_as_dict['class']}"
        result = mqtt_client.publish(
            topic=topic,
            payload=next_to_send.payload,
            qos=0,
            retain=False,
            properties=None
        )
        client_logger.debug(
            f"publish returned rc: {result.rc}")

    # Have to let the last message drain before existing
    if result is not None:
        loop = asyncio.get_event_loop()
        t = loop.call_later(1, exit, 1)
        result.wait_for_publish()
        logger.info("Published, bye!")
        t.cancel()








"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only

On Linux, Bluetooth devices are not "cleaned up" by the OS when a process exits.
Although both the bleak library and this code attempts to ensure that they
are not left in a connected state by the OS's BlueZ stack, there are conditions
when this may still happen, such as SIGKILL or low-level faults. On relaunch,
these devices generally will not be advertising (as they are connected) and
can't be connected.

These functions manage files in config.bluetooth.ID_FILE_DIRECTORY, suffixed by
config.bluetooth.ID_FILE_SUFFIX, that may be used by a supervisory script
to disconnect any that remain connected.

After extracting the device ID from the file and properly sanitizing it,

    bluetoothctl disconnect D9:B2:48:aa:bb:cc

can run by a member of the bluetooth group to disconnect the device.

"""

import os
import re
import sys
from pathlib import Path

from pyDE1.config import config

re_nonhex = re.compile('[^0-9a-fA-F]')


def filename_from_id(id: str) -> Path:
    if id is None:
        raise ValueError("Attempt to persist None as Bluetooth ID")
    suffix = config.bluetooth.ID_FILE_SUFFIX
    fname = re.sub(re_nonhex, '', id)
    # This is only "active" for Linux, so expect 12, hex characters
    if len(fname) != 12:
        raise ValueError(
            f"Hex-filtered ID '{fname}' from '{id}' is not 12 characters")
    if not suffix.startswith('.'):
        suffix = '.' + suffix
    return Path(config.bluetooth.ID_FILE_DIRECTORY, fname + suffix)


def persist_connection_file(id: str):
    if sys.platform != 'linux':
        return
    with open(filename_from_id(id), 'w') as fh:
        print(id, file=fh, end='')


def remove_connection_file(id: str):
    if sys.platform != 'linux':
        return
    try:
        os.remove(filename_from_id(id))
    except FileNotFoundError:
        pass



"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only
"""

import asyncio
import enum

from pyDE1.exceptions import DE1APIError


class CUUID (enum.Enum):
    """
    Maps the various C API objects to their corresponding BLE CUUIDs.

    A CUUID itself should be sufficient as a parameter in most cases.
    When needing the full UUID string, use CUUID.uuid

    A lock is provided, associated with each CUUID in case exclusive access
    is found to be needed. It is presently used by
    de1.write_packed_attr_return_notification()
    """
    Versions =          "a001"  # read,        notify
    RequestedState =    "a002"  # read, write, notify
    SetTime =           "a003"  # read, write, notify
    ShotDirectory =     "a004"  # read, write, notify
    ReadFromMMR =       "a005"  # read, write, notify
    WriteToMMR =        "a006"  #       write, notify
    ShotMapRequest =    "a007"  #       write, notify
    DeleteShotRange =   "a008"  #       write, notify
    FWMapRequest =      "a009"  #       write, notify
    Temperatures =      "a00a"  # read,        notify
    ShotSettings =      "a00b"  # read, write, notify
    Deprecated =        "a00c"  # read, write, notify
    ShotSample =        "a00d"  # read,        notify
    StateInfo =         "a00e"  # read,        notify
    HeaderWrite =       "a00f"  # read, write, notify
    FrameWrite =        "a010"  # read, write, notify
    WaterLevels =       "a011"  # read, write, notify
    Calibration =       "a012"  # read, write, notify

    def __init__(self, value):
        # super(CUUID, self).__init__(value)
        #     NameError: name 'CUUID' is not defined
        enum.Enum.__init__(value)
        # TODO: Confirm that locking by CUUID is sufficient
        # TODO: Decide if locking the "profile" is sufficient
        self._lock = asyncio.Lock()

    @property
    def can_read(self) -> bool:
        """
        Hard-coded test of the CUUID supports read.

        Does not take into account if the DE! is connected or not.

        The ReadFromMMR requests are in words, not bytes
        They are not properly recognized at this time by ReadFromMMR_callback
        For now, declare as not readable
        """
        return not (self in (CUUID.ReadFromMMR,
                             #
                             CUUID.WriteToMMR,
                             CUUID.ShotMapRequest,
                             CUUID.DeleteShotRange,
                             CUUID.FWMapRequest,
                             )
                    )

    @property
    def can_write(self) -> bool:
        """
        Hard-coded test of the CUUID supports write.

        Does not take into account if the DE! is connected or not.
        """
        return not (self in (CUUID.Versions,
                             CUUID.Temperatures,
                             CUUID.ShotSample,
                             CUUID.StateInfo,
                             )
                    )

    @property
    def can_notify(self) -> bool:
        """
        Hard-coded test of the CUUID supports notify.

        All are listed as notifying, though some do not return anything.

        Does not take into account if the DE! is connected or not.
        """
        return True

    # Very conservative right now
    # Even if a notify comes, if the command is "bad"
    # nothing may ever come back
    # Not FrameWrite as it does not notify on write
    @property
    def can_write_then_return(self) -> bool:
        """
        Hard-coded test of the CUUID supports write and will then notify.

        "Unrecognized" commands written may not result in a response,
        so waiting on a response may "hang" as a result.

        Does not take into account if the DE1 is connected or not.
        """
        return self in (
            CUUID.FWMapRequest,
            CUUID.Calibration,
        )

    @property
    def uuid(self) -> str:
        return f"0000{self.value}-0000-1000-8000-00805f9b34fb"

    @property
    def lock(self) -> asyncio.Lock:
        return self._lock

    @property
    def is_read_once(self):
        """
        Those that don't change with time (except at reboot)
        """
        return self is self.Versions

    @property
    def is_stable(self):
        """
        Those that don't change without being written
        """
        return self.is_read_once or self is self.Calibration


class UnsupportedBLEActionError(DE1APIError):
    def __init__(self, *args, **kwargs):
        super(UnsupportedBLEActionError, self).__init__(args, kwargs)


"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only

Events that are shared across multiple classes

See also:
    EventWithNotify that wraps asyncio.Event()
    async def send_to_outbound_queue(payload: EventPayload)
"""

import enum
from typing import Optional

from pyDE1.event_manager import EventPayload


class ConnectivityState (enum.Enum):
    UNKNOWN = 'unknown'
    CONNECTING = 'connecting'
    CONNECTED = 'connected'
    READY = 'ready'  # "Ready for use"
    NOT_READY = 'not_ready'  # Was READY, but is no longer
    DISCONNECTING = 'disconnecting'
    DISCONNECTED = 'disconnected'


class ConnectivityChange (EventPayload):

    def __init__(self,
                 arrival_time: float,
                 state: ConnectivityState = ConnectivityState.UNKNOWN,
                 id: Optional[str] = None,
                 name: Optional[str] = None,
                 ):
        super(ConnectivityChange, self).__init__(arrival_time=arrival_time)
        self._version = "1.1.0"
        self.state = state
        self.id = id
        self.name = name


class FirmwareUploadState (enum.Enum):
    STARTING = 'starting'
    UPLOADING = 'uploading'
    COMPLETED = 'completed'
    FAILED = 'failed'
    CANCELED = 'canceled'


class FirmwareUpload (EventPayload):

    def __init__(self,
                 arrival_time: float,
                 state: FirmwareUploadState,
                 uploaded: Optional[int] = None,
                 total: Optional[int] = None,
                 ):
        super(FirmwareUpload, self).__init__(arrival_time=arrival_time)
        self._version = "1.0.0"
        self.state = state
        self.uploaded = uploaded
        self.total = total


"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only
"""

import asyncio
import enum
import inspect
import time
from typing import (
    Optional, Iterable, NamedTuple, Dict, Set, Tuple, Union,
)

from bleak import BleakScanner
from bleak.backends.device import BLEDevice
from bleak.backends.scanner import AdvertisementData

import pyDE1
from pyDE1.config import config
from pyDE1.event_manager import EventPayload, send_to_outbound_pipes
from pyDE1.exceptions import *
from pyDE1.singleton import Singleton
from pyDE1.supervise import SupervisedTask

logger = pyDE1.getLogger('Bluetooth.Scanner')

# Otherwise will iterate over D,E,1
_registered_ble_prefixes = set()
_registered_ble_prefixes.add('DE1')


class BleakScannerWrapped (BleakScanner):

    def __init__(self, **kwargs):
        super(BleakScannerWrapped, self).__init__(**kwargs)
        self._run_ref = 0

    @property
    def run_id(self):
        """
        This should be considered as opaque by consumers
        """
        return f"{self.__class__.__qualname__}_0x{id(self):x}_{self._run_ref}"

    async def start(self):
        self._run_ref += 1
        logger.debug("Starting scanner")
        ep = ScannerNotification(action=ScannerNotificationAction.STARTED,
                                 run_id=self.run_id)
        await send_to_outbound_pipes(ep)
        await super(BleakScannerWrapped, self).start()

    async def stop(self):
        await super(BleakScannerWrapped, self).stop()
        ep = ScannerNotification(action=ScannerNotificationAction.ENDED,
                                 run_id=self.run_id)
        await send_to_outbound_pipes(ep)
        logger.debug("Stopped scanner")


class ScannerNotificationAction (enum.Enum):
    STARTED = 'started'
    ENDED = 'ended'
    FOUND = 'found'


class ScannerNotification (EventPayload):
    """
    As these don't go through a normal "publish" but are only over MQTT,
    supply a unique ID for this scanning run in run_id so the recipient
    can associate started, found, and ended with a single "run"
    """
    def __init__(self, action: ScannerNotificationAction,
                 run_id: Optional[str] = None,
                 id: Optional[str] = None,
                 name: Optional[str] = None,
                 arrival_time: Optional[float] = None,
                 create_time: Optional[float] = None):
        """
        STARTED and ENDED require None for id and name
        FOUND requires both id and name, though name potentially could be ''
        """
        if arrival_time is None:
            arrival_time = time.time()
        if action in (
                ScannerNotificationAction.STARTED,
                ScannerNotificationAction.ENDED) \
                and (id is not None or name is not None):
            raise DE1TypeError (
                "STARTED and ENDED require None for both id and name")
        if action is ScannerNotificationAction.FOUND \
                and (id is None or name is None):
            raise DE1TypeError (
                "FOUND requires id and name")
        super(ScannerNotification, self).__init__(arrival_time, create_time)
        self._version = '1.0.0'
        self.run_id = run_id
        self.action = action
        self.id = id
        self.name = name


async def notify_bledevice(device: BLEDevice, run_id: Optional[str] = None):
    for prefix in _registered_ble_prefixes:
        if device.name.startswith(prefix):
            ep = ScannerNotification(action=ScannerNotificationAction.FOUND,
                                     id=device.address,
                                     name=device.name,
                                     run_id=run_id)
            asyncio.create_task(send_to_outbound_pipes(ep))
            break


class DiscoveredDeviceEntry (NamedTuple):
    last_seen: float
    device: BLEDevice


class DiscoveredDevices (Singleton):

    def _singleton_init(self, *args, **kwds):
        self._devices_seen: Dict[str, DiscoveredDeviceEntry] = dict()
        self._lock = asyncio.Lock()
        self._queue = asyncio.Queue()
        self._worker = SupervisedTask(self._device_add_queue_worker)

    def add(self, device: BLEDevice, run_id: Optional['str'] = None):
        self._queue.put_nowait(self.AddThis(device=device,
                                            seen_at=time.time(),
                                            run_id=run_id))

    async def devices_seen(self,
                           starts_with_set: Optional[Iterable[str]] = None,
                           bledevice_only=False) \
            -> Set[Union[DiscoveredDeviceEntry, BLEDevice]]:
        # This is going to "lose" active devices as they time out
        async with self._lock:
            retval = set()
            now = time.time()
            pruned = dict()
            for addr, entry in self._devices_seen.items():
                if config.bluetooth.SCAN_CACHE_EXPIRY is not None \
                        and (now - entry.last_seen) \
                            > config.bluetooth.SCAN_CACHE_EXPIRY:
                    pass  # Can't delete in place during iteration
                else:
                    pruned[addr] = entry
            self._devices_seen = pruned

            for addr, entry in self._devices_seen.items():
                include = False
                if starts_with_set is not None:
                    for prefix in starts_with_set:
                        if entry.device.name.startswith(prefix):
                            include = True
                            break
                else:
                    include = True

                if include:
                    if bledevice_only:
                        retval.add(entry.device)
                    else:
                        retval.add(entry)
        return retval

    async def devices_for_json(self):
        retval = list()
        for entry in await self.devices_seen(
                starts_with_set=_registered_ble_prefixes,
                bledevice_only=False):
            retval.append({
                'id': entry.device.address,
                'name': entry.device.name,
                'discovered': entry.last_seen
            })
        return retval

    def ble_device_from_id(self, id: str) -> Optional[BLEDevice]:
        try:
            ble_device = self._devices_seen[id].device
        except KeyError:
            ble_device = None
        return ble_device

    async def clear(self):
        async with self._lock:
            self._devices_seen.clear()

    async def _device_add_queue_worker(self):
        while True:
            add_this = await self._queue.get()
            # Filtering is done in notify_bledevice
            asyncio.create_task(notify_bledevice(device=add_this.device,
                                                 run_id=add_this.run_id))
            async with self._lock:
                self._devices_seen[add_this.device.address] \
                    = DiscoveredDeviceEntry(last_seen=add_this.seen_at,
                                            device=add_this.device)

    class AddThis (NamedTuple):
        device: BLEDevice
        seen_at: float
        run_id: str


async def stop_scanner_if_running(scanner: BleakScanner):
    logger = pyDE1.getLogger('Bluetooth.Scanner')
    logger.debug(f"Stopping {scanner}")
    await scanner.stop()
    logger.debug(f"Stopped {scanner}")
    # try:
    #     scanning = scanner.is_scanning
    # except AttributeError:  # On Linux
    #     scanning = None
    # if scanning:
    #     logger.info(f"Is scanning")
    #     await scanner.stop()
    # elif scanning is None:
    #     logger.info(f"is_scanning returned None")
    #     try:
    #         await scanner.stop()
    #         logger.info(f"Scanner stopped")
    #     except KeyError:
    #         logger.info("Ignoring KeyError on scanner stop")
    # else:
    #     logger.info(f"NOT scanning")


async def find_first_matching(prefix_set: Iterable[str],
                              timeout=None) -> BLEDevice:

    if timeout is None:
        timeout = config.bluetooth.SCAN_TIME

    if isinstance(prefix_set, str):
        raise DE1TypeError(
            "Prefix set of type str will iterate the characters. "
            "('DE1',) is better")

    # scanner = BleakScannerWrapped()
    dd = DiscoveredDevices()

    def is_match(device: BLEDevice, adv: AdvertisementData) -> bool:
        nonlocal dd
        # NB: This is really, really fragile
        # [2] bleak/backends/bluezdbus/scanner.py',
        #   line 276, code _invoke_callback
        # [3] bleak/backends/bluezdbus/scanner.py',
        #   line 349, code _parse_msg
        run_id = None
        try:
            run_id = \
                inspect.currentframe().f_back.f_back.f_locals['self'].run_id
        finally:
            pass
        dd.add(device=device, run_id=run_id)

        retval = False
        for prefix in prefix_set:
            if adv.local_name and adv.local_name.startswith(prefix):
                logger.info(
                    f"'{prefix}' matched at {device.address} "
                    f"by {adv}")
                retval = True
        return retval

    # Using the .find_first_device_by_filter() function creates
    # an instance of cls, so don't have access to it for ID
    return await BleakScannerWrapped.find_device_by_filter(filterfunc=is_match)


async def scan_until_timeout(timeout=None) -> Tuple[BleakScannerWrapped,
                                                    float,
                                                    asyncio.Event]:
    """
    Returns (scanner, timeout, event) where event will be set
    once timeout expires and scanner stops
    """

    if timeout is None:
        timeout = config.bluetooth.SCAN_TIME

    scanner = BleakScannerWrapped()
    event = asyncio.Event()
    dd = DiscoveredDevices()

    def add_to_dd(device: BLEDevice, adv: AdvertisementData) -> None:
        nonlocal dd, scanner
        dd.add(device=device, run_id=scanner.run_id)

    scanner.register_detection_callback(add_to_dd)

    async def stop_later(after: float):
        nonlocal scanner, event
        await asyncio.sleep(after)
        await scanner.stop()
        event.set()

    await scanner.start()
    asyncio.create_task(stop_later(timeout))

    return scanner, timeout, event


async def scan_from_api(doit: bool):
    if doit:
        (scanner, timeout, event) = await scan_until_timeout()
        return {
            'run_id': scanner.run_id,
            'timeout': timeout
        }

    else:
        return None


"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only

Initial thoughts on EventManager

Should be able to dispatch quickly, so probably should create tasks
and add them to the loop

A subscriber should be able to remove itself. This supports one-shot
as well as being able to do coarse-grained, state-based filtering.
"Only send me updates if recording a shot." Need some kind of handle
on the subscription. Return a uuid.uuid4() ?

The event_data should always contain:
  arrival_time -- the time at which the "triggering action" occurred
  event_time -- the time that the event was told to dispatch
  sender -- the high-level object that generated the event,
            de1, scale, scale_processor, BLE changes, API, ...
  data -- an object containing data associated with the event, can be None

The list of available events should be discoverable, along with a description
of what they are. Is there a way to get all the defined subclasses of a class?
pyDE1.de1.c_api.PackedAttr.__subclasses__()

There should be a queue, in case things back up
Entries in the queue should be tagged with insertion time as event_time
This can then be compared with delivery time, as well as by a consumer

A queue with a single worker will suggest, but not guarantee, ordered delivery

One queue per event type seems a good start, though, for example,
"on every state change" might be processed before or after "on flow change"
One way to resolve this is to only notify once for anything.
(Yes, potentially use flags for the DE1 state info, major, minor,
flow state enter, flow state leave, flow phase change)

Use a mask to filter? Maybe later -- state changes, but what else?

What happens if the worker dies? How does this get detected?
Watchdog? Check on every add?

Don't need a queue to start as there will only be one publisher for an event.
"""

import asyncio
import multiprocessing
import multiprocessing.connection as mpc
import queue
import time
import uuid
from inspect import iscoroutinefunction, signature
from typing import Optional, Coroutine, Union, List

import pyDE1
from pyDE1.event_manager.payloads import (
    EventNotification, EventNotificationAction, EventNotificationName,
    EventPayload, SequencerGateName, SequencerGateNotification
)
from pyDE1.flow_sequencer import FlowSequencer

logger = pyDE1.getLogger('EventManager')

LOG_DELAYS = False


# Extend asyncio.Event() to notify over the API as well


class EventWithNotify (asyncio.Event):

    def __init__(self, sender, name: EventNotificationName):
        if not isinstance(name, EventNotificationName):
            raise TypeError(
                "EventNotification needs a valid EventNotificationName, not "
                f"'{name}'")
        super(EventWithNotify, self).__init__()
        self._sender = sender
        self._event_name = name

    def set(self):
        super(EventWithNotify, self).set()
        en = EventNotification(
            arrival_time=time.time(),
            sender=self._sender,
            name=self._event_name,
            action=EventNotificationAction.SET
        )
        asyncio.create_task(send_to_outbound_pipes(en))

    def clear(self):
        super(EventWithNotify, self).clear()
        en = EventNotification(
            arrival_time=time.time(),
            sender=self._sender,
            name=self._event_name,
            action=EventNotificationAction.CLEAR
        )
        asyncio.create_task(send_to_outbound_pipes(en))


class SequencerGate (EventWithNotify):
    """
    Extends to add the sequence ID
    """

    def __init__(self, sender, name: SequencerGateName):
        if not isinstance(name, SequencerGateName):
            raise TypeError(
                "EventNotification needs a valid SequencerGateName, not "
                f"'{name}'")
        super(SequencerGate, self).__init__(sender=sender, name=name)

    def set(self):
        # Get to the Event without sending another notification
        super(EventWithNotify, self).set()
        sgn = SequencerGateNotification(
            arrival_time=time.time(),
            sender=self._sender,
            name=self._event_name,
            action=EventNotificationAction.SET
        )
        # Here's where active_state gets set
        self._sender: FlowSequencer
        sgn.active_state = self._sender.active_state
        asyncio.create_task(send_to_outbound_pipes(sgn))

    def clear(self):
        super(EventWithNotify, self).clear()
        sgn = SequencerGateNotification(
            arrival_time=time.time(),
            sender=self._sender,
            name=self._event_name,
            action=EventNotificationAction.CLEAR
        )
        # Here's where active_state gets set
        self._sender: FlowSequencer
        sgn.active_state = self._sender.active_state
        asyncio.create_task(send_to_outbound_pipes(sgn))


async def send_to_outbound_pipes(payload: EventPayload):

    # multiprocessing.queue() can block at least on "full"
    # macOS doesn't support qsize()

    if payload._event_time is None:
        payload._event_time = time.time()
    q_payload = payload.as_json()
    # for pipe in [SubscribedEvent.outbound_pipe,
    #              SubscribedEvent.database_queue]:
    #     if SubscribedEvent.outbound_pipe is not None:
    #         pipe.send(q_payload)
    SubscribedEvent.outbound_pipe.send(q_payload)
    # Will raise queue.Full
    try:
        SubscribedEvent.database_queue.put_nowait(q_payload)
    except queue.Full:
        logger.error("Database queue full")
        # Not much else to do. Killing a shot in progress seems excessive


class SubscribedEvent:
    """
    A canonical pub/sub system for EventPayload objects under asyncio

    This is intended to be a singleton for each event type
    and with a single publisher.

    If outbound_pipe: is not None
    and EventPayload._internal_only is False (default)
    the EventPayload.as_json() will be delivered to that queue.

    SubscribedEvent.outbound_pipe and EventPayload._internal_only
    are class properties at this time.

    """

    outbound_pipe: Optional[mpc.Connection] = None
    database_queue: Optional[multiprocessing.Queue] = None

    def __init__(self, sender):
        self._sender = sender
        self._subscribers = []
        # perhaps don't need a lock on the list as not threaded
        self._subscriber_list_lock = asyncio.Lock()
        self._last_create_time = 0

    async def subscribe(self,
                        callback: Coroutine[
                            None, EventPayload, None]) -> uuid.UUID:
        """
        Subscribe to the series of events

        Returns a UUID that can be later used to unsubscribe
        """
        # iscoroutine() fails on "async def something(payload)
        if not iscoroutinefunction(callback):
            raise TypeError(
                f"The callback must be a coroutine function: {callback}")

        # For now, just assume that optional parameters aren't being used
        scb = signature(callback)
        if len(scb.parameters) != 1:
            raise TypeError(
                f"The callback must accept a single argument: {scb.parameters}")

        id = uuid.uuid4()
        async with self._subscriber_list_lock:
            self._subscribers.append((id, callback))
        logger.debug(f"Subscribed {callback} to {self}")
        return id

    async def unsubscribe(self, id: Union[uuid.UUID, str,
                                          None]) -> Union[bool, None]:
        """
        Unsubscribe from an event
        Returns True on success, None if not found, and False on failure
        (presently that the id isn't a valid UUID')

        If None is the id, it returns True
        (allows for removal of placeholders that haven't been subscribed)
        """
        if id is None:
            return True
        retval = None
        if isinstance(id, str):
            try:
                id = uuid.UUID(id)
            except ValueError:
                return False
        async with self._subscriber_list_lock:
            len_before = len(self._subscribers)
            self._subscribers = list(
                filter(lambda s: s[0] != id, self._subscribers))
            len_after = len(self._subscribers)
        if len_after < len_before:
            retval = True
        else:
            retval = None
        return retval

    async def publish(self, payload: EventPayload) -> List[asyncio.Task]:
        """
        Take an EventPayload and distribute to all subscribers

        Returns a list of tasks created

        TODO: Consider if a deepcopy of the EventPayload is justified
              (It probably is, as it gets passed by reference)
              (Or not, as it isn't exposed as a writable object through an API)
        """
        payload._sender = self._sender
        async with self._subscriber_list_lock:
            payload._event_time = time.time()
            tasks = []
            for s in self._subscribers:
                # These have ben validated as coroutines
                # with single arguments on subscribe()
                #
                # # TODO: The copy() isn't being done asynchronously
                #
                # # TODO: Check this with the profiler
                # #       Might be better to off the whole thing
                # #       await/gather is scary if one hangs
                #
                # t = asyncio.create_task(s[1](copy(payload)))
                # tasks.append(t)
                # await s[1](copy(payload))

                await s[1](payload)
        internal_done = time.time()

        # multiprocessing.queue() can block at least on "full"
        # macOS doesn't support qsize()
        if not payload._internal_only:
            await send_to_outbound_pipes(payload)

        delivery_done = time.time()

        if LOG_DELAYS and not payload._internal_only:
            logger.info(
                "JSON and queueing time: "
                f"{(delivery_done - internal_done) * 1000:0.3f} ms")

        if (interpacket := payload.create_time - self._last_create_time) < 0:
            logger.error(
                "Out-of-sequence payload delivery by "
                f"{-interpacket * 1000:0.3f} ms")
        if LOG_DELAYS:
            delay = (delivery_done - payload.create_time) * 1000
            logger.debug( f"Dispatch delay: {delay:.3f} ms {type(payload)}")
        return tasks



"""
Copyright © 2021 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only

Common events for scales
"""

from pyDE1.event_manager import EventPayload


class ScaleWeightUpdate(EventPayload):
    """
    See WeightAndFlowUpdate for API-visible class
    """
    _internal_only = True

    def __init__(self,
                 arrival_time: float,
                 scale_time: float,
                 weight: float, ):
        super(ScaleWeightUpdate, self).__init__(arrival_time=arrival_time)
        self._version = "1.0.0"
        self.scale_time = scale_time
        self.weight = weight


class ScaleButtonPress(EventPayload):

    def __init__(self, arrival_time: float, button: int):
        super(ScaleButtonPress, self).__init__(arrival_time=arrival_time)
        self._version = "1.0.0"  # Major version incremented on breaking change
        self.button = button


class ScaleTareSeen(EventPayload):

    def __init__(self, arrival_time: float):
        super(ScaleTareSeen, self).__init__(arrival_time=arrival_time)
        self._version = "1.0.0"  # Major version incremented on breaking change


class WeightAndFlowUpdate(EventPayload):
    """
    On ScaleProcessor at this time
    """
    # Right now, this doesn't capture any DE1 data, such as state
    # I'm not sure that it needs to with this framework
    def __init__(self, arrival_time: float,
                 scale_time: float,
                 current_weight: float,
                 current_weight_time: float,
                 average_flow: float,
                 average_flow_time: float,
                 median_weight: float,
                 median_weight_time: float,
                 median_flow: float,
                 median_flow_time: float,
                 ):
        super(WeightAndFlowUpdate, self).__init__(arrival_time=arrival_time)
        self._version = "1.0.0"  # Major version incremented on breaking change
        self.scale_time = scale_time
        self.current_weight: float = current_weight
        self.current_weight_time: float = current_weight_time
        self.average_flow: float = average_flow
        self.average_flow_time: float = average_flow_time
        self.median_weight: float = median_weight
        self.median_weight_time: float = median_weight_time
        self.median_flow: float = median_flow
        self.median_flow_time: float = median_flow_time
